from PyQt5.QtWidgets import QMessageBox, QScrollArea, QButtonGroup, QComboBox, \
    QCheckBox, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton, \
    QRadioButton, QSizePolicy
from PyQt5.QtCore import Qt, QRect
from PyQt5.QtGui import QIcon, QDoubleValidator

from celldetective.gui import Styles
from celldetective.gui.gui_utils import center_window, FigureCanvas

from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
from celldetective.utils import get_software_location, _extract_labels_from_config
from celldetective.io import load_experiment_tables, get_experiment_antibodies, get_experiment_cell_types, get_experiment_concentrations, \
    get_positions_in_well, get_experiment_wells
from celldetective.signals import mean_signal
import numpy as np
import json
import os
import matplotlib.pyplot as plt

plt.rcParams['svg.fonttype'] = 'none'
from glob import glob
import pandas as pd
from matplotlib.cm import tab10
import math
import seaborn as sns


class ConfigMeasurementsPlot(QWidget,Styles):
    """
    UI to set survival instructions.

    """

    def __init__(self, parent_window=None):

        super().__init__()
        self.parent_window = parent_window
        self.setWindowTitle("Configure signal plot")
        self.setWindowIcon(QIcon(os.sep.join(['celldetective', 'icons', 'mexican-hat.png'])))
        self.exp_dir = self.parent_window.exp_dir
        self.soft_path = get_software_location()
        self.exp_config = self.exp_dir + "config.ini"
        self.wells = np.array(self.parent_window.parent_window.wells, dtype=str)
        self.well_labels = _extract_labels_from_config(self.exp_config, len(self.wells))
        self.FrameToMin = self.parent_window.parent_window.FrameToMin
        self.float_validator = QDoubleValidator()
        self.target_class = [0, 1]
        self.show_ci = True
        self.show_cell_lines = False
        self.ax2 = None
        self.auto_close = False
        self.palette = sns.cubehelix_palette()
        # sns.color_palette("cubehelix", as_cmap=True)
        # sns.color_palette("ch:s=-.2,r=.6", as_cmap=True)

        print('Parent wells: ', self.wells)

        self.well_option = self.parent_window.parent_window.well_list.currentIndex()
        self.position_option = self.parent_window.parent_window.position_list.currentIndex()
        self.interpret_pos_location()
        # self.load_available_tables()
        # self.config_path = self.exp_dir + self.config_name

        self.screen_height = self.parent_window.parent_window.parent_window.screen_height
        center_window(self)
        # self.setMinimumHeight(int(0.8*self.screen_height))
        # self.setMaximumHeight(int(0.8*self.screen_height))
        self.populate_widget()
        # self.load_previous_measurement_instructions()
        if self.auto_close:
            self.close()
        self.hue_colors = {0: self.palette[0], 1: self.palette[3]}

    def interpret_pos_location(self):

        """
        Read the well/position selection from the control panel to decide which data to load
        Set position_indices to None if all positions must be taken

        """

        if self.well_option == len(self.wells):
            self.well_indices = np.arange(len(self.wells))
        else:
            self.well_indices = np.array([self.well_option], dtype=int)

        if self.position_option == 0:
            self.position_indices = None
        else:
            self.position_indices = np.array([self.position_option], dtype=int)

    def populate_widget(self):

        """
        Create the multibox design.

        """

        # Create button widget and layout
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(30, 30, 30, 30)
        panel_title = QLabel('Options')
        panel_title.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
        main_layout.addWidget(panel_title, alignment=Qt.AlignCenter)

        labels = [QLabel('population: '), QLabel('class: '), QLabel('group: ')]  # , QLabel('time of\ninterest: ')]
        self.cb_options = [['targets', 'effectors'], ['class'], ['group']]  # , ['t0','first detection']]
        self.cbs = [QComboBox() for i in range(len(labels))]
        self.cbs[0].currentIndexChanged.connect(self.set_classes_and_times)

        choice_layout = QVBoxLayout()
        choice_layout.setContentsMargins(20, 20, 20, 20)
        for i in range(len(labels)):
            hbox = QHBoxLayout()
            hbox.addWidget(labels[i], 33)
            hbox.addWidget(self.cbs[i], 66)
            self.cbs[i].addItems(self.cb_options[i])
            if i == 1:
                vbox = QVBoxLayout()
                self.check_class = QCheckBox("plot by class")
                vbox.addWidget(self.check_class)
                vbox.addLayout(hbox)
                choice_layout.addLayout(vbox)
            if i == 2:
                vbox = QVBoxLayout()
                self.check_group = QCheckBox("plot by group")
                vbox.addWidget(self.check_group)
                vbox.addLayout(hbox)
                choice_layout.addLayout(vbox)
            else:
                choice_layout.addLayout(hbox)

        self.cbs[0].setCurrentIndex(1)
        self.cbs[0].setCurrentIndex(0)
        self.cbs[1].setEnabled(False)
        self.check_class.toggled.connect(self.class_enable)
        self.cbs[2].setEnabled(False)
        self.check_group.toggled.connect(self.group_enable)

        # self.abs_time_checkbox = QCheckBox('absolute time')
        # self.frame_slider = QLabeledSlider()
        # self.frame_slider.setSingleStep(1)
        # self.frame_slider.setOrientation(1)
        # self.frame_slider.setRange(0,self.parent.parent.len_movie)
        # self.frame_slider.setValue(0)
        # self.frame_slider.setEnabled(False)
        # slider_hbox = QHBoxLayout()
        # slider_hbox.addWidget(self.abs_time_checkbox, 33)
        # slider_hbox.addWidget(self.frame_slider, 66)
        # choice_layout.addLayout(slider_hbox)
        main_layout.addLayout(choice_layout)

        # self.abs_time_checkbox.stateChanged.connect(self.switch_ref_time_mode)

        # time_calib_layout = QHBoxLayout()
        # time_calib_layout.setContentsMargins(20,20,20,20)
        # time_calib_layout.addWidget(QLabel('time calibration\n(frame to min)'), 33)
        # self.time_calibration_le = QLineEdit(str(self.FrameToMin).replace('.',','))
        # self.time_calibration_le.setValidator(self.float_validator)
        # time_calib_layout.addWidget(self.time_calibration_le, 66)
        # #time_calib_layout.addWidget(QLabel(' min'))
        # main_layout.addLayout(time_calib_layout)

        self.submit_btn = QPushButton('Submit')
        self.submit_btn.setStyleSheet(self.button_style_sheet)
        self.submit_btn.clicked.connect(self.process_signal)
        main_layout.addWidget(self.submit_btn)

    # self.populate_left_panel()
    # grid.addLayout(self.left_side, 0, 0, 1, 1)

    # self.setCentralWidget(self.scroll_area)
    # self.show()
    def class_enable(self):
        if self.check_class.isChecked():
            self.cbs[1].setEnabled(True)
            self.check_group.setChecked(False)
        else:
            self.cbs[1].setEnabled(False)

    def group_enable(self):
        if self.check_group.isChecked():
            self.cbs[2].setEnabled(True)
            self.check_class.setChecked(False)
        else:
            self.cbs[2].setEnabled(False)

    def set_classes_and_times(self):
        ext = 'csv'
        # Look for all classes and times
        tables = glob(self.exp_dir + os.sep.join(['W*', '*', 'output', 'tables', f'trajectories_*{ext}']))
        self.all_columns = []
        for tab in tables:
            cols = pd.read_csv(tab, nrows=1).columns.tolist()
            self.all_columns.extend(cols)
        self.all_columns = np.unique(self.all_columns)
        class_idx = np.array([s.startswith('class_') for s in self.all_columns])
        group_idx = np.array([s.startswith('group_') for s in self.all_columns])

        # time_idx = np.array([s.startswith('t_') for s in self.all_columns])

        try:
            class_columns = list(self.all_columns[class_idx])
            group_columns = list(self.all_columns[group_idx])
        # time_columns = list(self.all_columns[time_idx])
        except:
            print('columns not found')
            self.auto_close = True
            return None

        # self.cbs[2].clear()
        # self.cbs[2].addItems(np.unique(self.cb_options[2]+time_columns))

        self.cbs[1].clear()
        self.cbs[1].addItems(np.unique(self.cb_options[1] + class_columns))
        self.cbs[2].clear()
        self.cbs[2].addItems(np.unique(group_columns))

    def ask_for_feature(self):

        cols = np.array(list(self.df.columns))
        is_number = np.vectorize(lambda x: np.issubdtype(x, np.number))
        feats = cols[is_number(self.df.dtypes)]

        self.feature_choice_widget = QWidget()
        self.feature_choice_widget.setWindowTitle("Select numeric feature")
        layout = QVBoxLayout()
        self.feature_choice_widget.setLayout(layout)
        self.feature_cb = QComboBox()
        self.feature_cb.addItems(feats)
        hbox = QHBoxLayout()
        hbox.addWidget(QLabel('feature: '), 33)
        hbox.addWidget(self.feature_cb, 66)
        layout.addLayout(hbox)

        self.set_feature_btn = QPushButton('set')
        self.set_feature_btn.clicked.connect(self.compute_signals)
        layout.addWidget(self.set_feature_btn)
        self.feature_choice_widget.show()
        center_window(self.feature_choice_widget)

    def ask_for_features(self):

        cols = np.array(list(self.df.columns))
        is_number = np.vectorize(lambda x: np.issubdtype(x, np.number))
        feats = cols[is_number(self.df.dtypes)]

        self.feature_choice_widget = QWidget()
        self.feature_choice_widget.setWindowTitle("Select numeric feature")
        layout = QVBoxLayout()
        self.feature_choice_widget.setLayout(layout)
        self.feature_cb = QComboBox()
        self.feature_cb.addItems(feats)
        hbox = QHBoxLayout()
        hbox.addWidget(QLabel('feature: '), 33)
        hbox.addWidget(self.feature_cb, 66)
        # hbox.addWidget((QLabel('Plot two features')))
        layout.addLayout(hbox)
        self.checkBox_feature = QCheckBox(self.feature_choice_widget)
        self.checkBox_feature.setGeometry(QRect(170, 120, 81, 20))
        # self.checkBox_feature.setText('Plot two features')
        # #hbox.addWidget(self.checkBox_feature)
        # layout.addWidget(self.checkBox_feature, alignment=Qt.AlignCenter)
        # self.checkBox_feature.stateChanged.connect(self.enable_second_feature)

        # layout = QVBoxLayout()
        # self.feature_two_choice_widget.setLayout(layout)
        # self.feature_two_cb = QComboBox()
        # self.feature_two_cb.addItems(feats)
        # self.feature_two_cb.setEnabled(False)
        # hbox_two = QHBoxLayout()
        # hbox_two.addWidget(QLabel('feature two: '), 33)
        # hbox_two.addWidget(self.feature_two_cb, 66)
        # layout.addLayout(hbox_two)

        self.set_feature_btn = QPushButton('set')
        self.set_feature_btn.clicked.connect(self.compute_signals)
        layout.addWidget(self.set_feature_btn)
        self.feature_choice_widget.show()
        center_window(self.feature_choice_widget)

    def enable_second_feature(self):
        if self.checkBox_feature.isChecked():
            self.feature_two_cb.setEnabled(True)
        else:
            self.feature_two_cb.setEnabled(False)

    def compute_signals(self):

        if self.df is not None:
            self.feature_selected = self.feature_cb.currentText()
            self.feature_choice_widget.close()
            self.compute_signal_functions()
            # prepare survival

            # plot survival
            self.survivalWidget = QWidget()
            self.scroll = QScrollArea()
            self.survivalWidget.setMinimumHeight(int(0.8 * self.screen_height))
            self.survivalWidget.setWindowTitle('signals')
            self.plotvbox = QVBoxLayout(self.survivalWidget)
            self.plotvbox.setContentsMargins(30, 30, 30, 30)
            self.survival_title = QLabel('Signal function')
            self.survival_title.setStyleSheet("""
				font-weight: bold;
				padding: 0px;
				""")
            self.plotvbox.addWidget(self.survival_title, alignment=Qt.AlignCenter)

            plot_buttons_hbox = QHBoxLayout()
            plot_buttons_hbox.addWidget(QLabel(''), 80, alignment=Qt.AlignLeft)

            self.legend_btn = QPushButton('')
            self.legend_btn.setIcon(icon(MDI6.text_box, color="black"))
            self.legend_btn.setStyleSheet(self.button_select_all)
            self.legend_btn.setToolTip('Show or hide the legend')
            self.legend_visible = True
            self.legend_btn.clicked.connect(self.show_hide_legend)
            plot_buttons_hbox.addWidget(self.legend_btn, 5, alignment=Qt.AlignRight)

            self.log_btn = QPushButton('')
            self.log_btn.setIcon(icon(MDI6.math_log, color="black"))
            self.log_btn.setStyleSheet(self.button_select_all)
            self.log_btn.clicked.connect(self.switch_to_log)
            self.log_btn.setToolTip('Enable or disable log scale')
            plot_buttons_hbox.addWidget(self.log_btn, 5, alignment=Qt.AlignRight)

            # self.ci_btn = QPushButton('')
            # self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal,color="blue"))
            # self.ci_btn.setStyleSheet(self.parent.parent.parent.button_select_all)
            # self.ci_btn.clicked.connect(self.switch_ci)
            # self.ci_btn.setToolTip('Show or hide confidence intervals.')
            # plot_buttons_hbox.addWidget(self.ci_btn, 5, alignment=Qt.AlignRight)

            self.cell_lines_btn = QPushButton('')
            self.cell_lines_btn.setIcon(icon(MDI6.view_headline, color="black"))
            self.cell_lines_btn.setStyleSheet(self.button_select_all)
            self.cell_lines_btn.clicked.connect(self.switch_cell_lines)
            self.cell_lines_btn.setToolTip('Show or hide individual cell signals.')
            plot_buttons_hbox.addWidget(self.cell_lines_btn, 5, alignment=Qt.AlignRight)

            self.fig, self.ax = plt.subplots(1, 1, figsize=(10, 10))  # ,figsize=(10,10)
            # self.setMinimumHeight(100)
            self.survival_window = FigureCanvas(self.fig, title="Survival")
            self.survival_window.setContentsMargins(0, 0, 0, 0)
            if self.df is not None:
                self.initialize_axis()
            plt.tight_layout()

            self.fig.set_facecolor('none')  # or 'None'
            self.fig.canvas.setStyleSheet("background-color: transparent;")
            self.survival_window.canvas.draw()

            # self.survival_window.layout.addWidget(QLabel('WHAAAAATTT???'))

            self.plot_options = [QRadioButton() for i in range(3)]
            self.radio_labels = ['well', 'pos', 'both']
            radio_hbox = QHBoxLayout()
            radio_hbox.setContentsMargins(30,30,30,30)
            self.plot_btn_group = QButtonGroup()
            for i in range(3):
            	self.plot_options[i].setText(self.radio_labels[i])
            	#self.plot_options[i].toggled.connect(self.plot_survivals)
            	self.plot_btn_group.addButton(self.plot_options[i])
            	radio_hbox.addWidget(self.plot_options[i], 33, alignment=Qt.AlignCenter)
            self.plot_btn_group.buttonClicked[int].connect(self.plot_survivals)

            if self.position_indices is not None:
                if len(self.well_indices) > 1 and len(self.position_indices) == 1:
                    self.plot_btn_group.buttons()[0].click()
                    for i in [1, 2]:
                        self.plot_options[i].setEnabled(False)
                elif len(self.well_indices) > 1:
                    self.plot_btn_group.buttons()[0].click()
                elif len(self.well_indices) == 1 and len(self.position_indices) == 1:
                    self.plot_btn_group.buttons()[1].click()
                    for i in [0, 2]:
                        self.plot_options[i].setEnabled(False)
            # else:
            # 	if len(self.well_indices)>1:
            # 		self.plot_btn_group.buttons()[0].click()
            # 	elif len(self.well_indices)==1:
            # 		self.plot_btn_group.buttons()[2].click()

            # elif len(self.well_indices)>1:
            # 	self.plot_btn_group.buttons()[0].click()
            # else:
            # 	self.plot_btn_group.buttons()[1].click()

            if self.position_indices is not None:
                for i in [0, 2]:
                    self.plot_options[i].setEnabled(False)

            # self.plot_options[0].setChecked(True)
            # self.plotvbox.addLayout(radio_hbox)
            self.plotvbox.addLayout(plot_buttons_hbox)
            self.plotvbox.addWidget(self.survival_window)
            # self.class_selection_lbl = QLabel('Select class')
            # self.class_selection_lbl.setStyleSheet("""
            # 	font-weight: bold;
            # 	padding: 0px;
            # 	""")
            # self.plotvbox.addWidget(self.class_selection_lbl, alignment=Qt.AlignCenter)
            # class_selection_hbox = QHBoxLayout()
            # class_selection_hbox.setContentsMargins(30,30,30,30)
            # self.all_btn = QRadioButton('*')
            # self.all_btn.setChecked(True)
            # self.event_btn = QRadioButton('event')
            # self.no_event_btn = QRadioButton('no event')
            # self.class_btn_group = QButtonGroup()
            # for btn in [self.all_btn, self.event_btn, self.no_event_btn]:
            # 	self.class_btn_group.addButton(btn)
            #
            # self.class_btn_group.buttonClicked[int].connect(self.set_class_to_plot)
            #
            # class_selection_hbox.addWidget(self.all_btn, 33, alignment=Qt.AlignLeft)
            # class_selection_hbox.addWidget(self.event_btn, 33, alignment=Qt.AlignCenter)
            # class_selection_hbox.addWidget(self.no_event_btn, 33, alignment=Qt.AlignRight)
            # self.plotvbox.addLayout(class_selection_hbox)

            self.select_pos_label = QLabel('Select positions')
            self.select_pos_label.setStyleSheet("""
				font-weight: bold;
				padding: 0px;
				""")
            self.plotvbox.addWidget(self.select_pos_label, alignment=Qt.AlignCenter)
            #
            # self.select_option = [QRadioButton() for i in range(2)]
            # self.select_label = ['name', 'spatial']
            # select_hbox = QHBoxLayout()
            # select_hbox.setContentsMargins(30,30,30,30)
            # self.select_btn_group = QButtonGroup()
            # for i in range(2):
            # 	self.select_option[i].setText(self.select_label[i])
            # 	#self.select_option[i].toggled.connect(self.switch_selection_mode)
            # 	self.select_btn_group.addButton(self.select_option[i])
            # 	select_hbox.addWidget(self.select_option[i],33, alignment=Qt.AlignCenter)
            # self.select_btn_group.buttonClicked[int].connect(self.switch_selection_mode)
            # self.plotvbox.addLayout(select_hbox)

            self.look_for_metadata()
            if self.metadata_found:
                self.fig_scatter, self.ax_scatter = plt.subplots(1, 1, figsize=(4, 3))
                self.position_scatter = FigureCanvas(self.fig_scatter)
                self.load_coordinates()
                self.plot_spatial_location()
                # self.plot_positions()
                self.ax_scatter.spines['top'].set_visible(False)
                self.ax_scatter.spines['right'].set_visible(False)
                self.ax_scatter.set_aspect('equal')
                self.ax_scatter.set_xticks([])
                self.ax_scatter.set_yticks([])
                plt.tight_layout()

                self.fig_scatter.set_facecolor('none')  # or 'None'
                self.fig_scatter.canvas.setStyleSheet("background-color: transparent;")
                self.plotvbox.addWidget(self.position_scatter)

            self.generate_pos_selection_widget()

            # if self.df is not None and len(self.ks_estimators_per_position)>0:
            # 	self.plot_survivals()
            # self.select_btn_group.buttons()[0].click()
            self.scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            # self.scroll.setWidgetResizable(True)
            self.scroll.setWidget(self.survivalWidget)

            self.scroll.setMinimumHeight(int(0.8 * self.screen_height))
            self.survivalWidget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self.scroll.show()

    def process_signal(self):

        print('you clicked!!')
        # self.FrameToMin = float(self.time_calibration_le.text().replace(',','.'))
        print(self.FrameToMin, 'set')

        # read instructions from combobox options
        self.load_available_tables()
        if self.df is not None:
            self.ask_for_features()
        else:
            return None

    def generate_pos_selection_widget(self):

        self.well_names = self.df['well_name'].unique()
        self.pos_names = self.df_pos_info[
            'pos_name'].unique()  # pd.DataFrame(self.ks_estimators_per_position)['position_name'].unique()
        print(f'POSITION NAMES: ', self.pos_names)
        self.usable_well_labels = []
        for name in self.well_names:
            for lbl in self.well_labels:
                if name + ':' in lbl:
                    self.usable_well_labels.append(lbl)

        self.line_choice_widget = QWidget()
        self.line_check_vbox = QVBoxLayout()
        self.line_choice_widget.setLayout(self.line_check_vbox)
        if len(self.well_indices) > 1:
            self.well_display_options = [QCheckBox(self.usable_well_labels[i]) for i in
                                         range(len(self.usable_well_labels))]
            for i in range(len(self.well_names)):
                self.line_check_vbox.addWidget(self.well_display_options[i], alignment=Qt.AlignLeft)
                self.well_display_options[i].setChecked(True)
                self.well_display_options[i].toggled.connect(self.select_survival_lines)
        else:
            self.pos_display_options = [QCheckBox(self.pos_names[i]) for i in range(len(self.pos_names))]
            for i in range(len(self.pos_names)):
                self.line_check_vbox.addWidget(self.pos_display_options[i], alignment=Qt.AlignLeft)
                self.pos_display_options[i].setChecked(True)
                self.pos_display_options[i].toggled.connect(self.select_survival_lines)

        self.plotvbox.addWidget(self.line_choice_widget, alignment=Qt.AlignCenter)

    def load_available_tables(self):

        """
        Load the tables of the selected wells/positions from the control Panel for the population of interest

        """

        self.well_option = self.parent_window.parent_window.well_list.currentIndex()
        if self.well_option == len(self.wells):
            wo = '*'
        else:
            wo = self.well_option
        self.position_option = self.parent_window.parent_window.position_list.currentIndex()
        if self.position_option == 0:
            po = '*'
        else:
            po = self.position_option - 1

        self.df, self.df_pos_info = load_experiment_tables(self.exp_dir, well_option=wo, position_option=po,
                                                           population=self.cbs[0].currentText(), return_pos_info=True)

        if self.df is None:

            print('No table could be found...')
            msgBox = QMessageBox()
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.setText("No table could be found to compute survival...")
            msgBox.setWindowTitle("Warning")
            msgBox.setStandardButtons(QMessageBox.Ok)
            returnValue = msgBox.exec()
            if returnValue == QMessageBox.Ok:
                self.close()
                return None
            else:
                self.close()
                return None
        else:
            self.df_well_info = self.df_pos_info.loc[:,
                                ['well_path', 'well_index', 'well_name', 'well_number', 'well_alias']].drop_duplicates()

    def compute_signal_functions(self):
        wells = get_experiment_wells(self.exp_dir)
        antibodies = get_experiment_antibodies(self.exp_dir)
        cell_types = get_experiment_cell_types(self.exp_dir)
        concentrations = get_experiment_concentrations(self.exp_dir)
        well_name = [' '.join([s1, '+', s2, s3, 'pM']) for s1, s2, s3 in zip(cell_types, antibodies, concentrations)]

        data = []
        full_data = []
        print(self.well_option)
        if self.well_option > 1:
            self.plot_mode = 'wells'
            for z, well in enumerate(wells):  # loop over wells
                # print(well)
                if z not in self.well_indices:
                    pass
                else:
                    positions = get_positions_in_well(well)
                    for ind, pos in enumerate(positions):  # loop over positions
                        if self.position_indices is not None:
                            if ind + 1 in self.position_indices:
                                print(f'Processing position {pos}...')
                                tab_tc = pos + os.sep.join(
                                    ['output', 'tables', f'trajectories_{self.cbs[0].currentText()}.csv'])
                                if not os.path.exists(tab_tc):
                                    return None
                                else:
                                    data = pd.read_csv(tab_tc)
                                    data['well_name'] = well_name[z]
                                full_data.append(data)

                        else:
                            print(f'Processing position {pos}...')
                            tab_tc = pos + os.sep.join(
                                ['output', 'tables', f'trajectories_{self.cbs[0].currentText()}.csv'])
                            if not os.path.exists(tab_tc):
                                return None
                            else:
                                data = pd.read_csv(tab_tc)
                                data['well_name'] = well_name[z]
                            full_data.append(data)
        else:
            self.plot_mode = 'positions'
            positions = get_positions_in_well(wells[self.well_indices[0]])
            for ind, pos in enumerate(positions):  # loop over positions
                pos_name = pos.split(os.sep)[-2]
                if self.position_indices is not None:
                    if ind + 1 in self.position_indices:
                        print(f'Processing position {pos}...')
                        tab_tc = pos + os.sep.join(
                            ['output', 'tables', f'trajectories_{self.cbs[0].currentText()}.csv'])
                        if not os.path.exists(tab_tc):
                            return None
                        else:
                            data = pd.read_csv(tab_tc)
                            data['position'] = pos_name
                        full_data.append(data)
                else:
                    print(f'Processing position {pos}...')
                    tab_tc = pos + os.sep.join(['output', 'tables', f'trajectories_{self.cbs[0].currentText()}.csv'])
                    if not os.path.exists(tab_tc):
                        return None
                    else:
                        data = pd.read_csv(tab_tc)
                        data['position'] = pos_name
                    full_data.append(data)
        self.plot_data = pd.concat(full_data, ignore_index=True)

    def generate_synchronized_matrix(self, well_group, feature_selected, cclass, max_time):

        if isinstance(cclass, int):
            cclass = [cclass]

        n_cells = len(well_group.groupby(['position', 'TRACK_ID']))
        depth = int(2 * max_time + 3)
        matrix = np.zeros((n_cells, depth))
        matrix[:, :] = np.nan
        mapping = np.arange(-max_time - 1, max_time + 2)
        cid = 0
        for block, movie_group in well_group.groupby('position'):
            for tid, track_group in movie_group.loc[movie_group[self.cbs[1].currentText()].isin(cclass)].groupby(
                    'TRACK_ID'):
                try:
                    timeline = track_group['FRAME'].to_numpy().astype(int)
                    feature = track_group[feature_selected].to_numpy()
                    if self.checkBox_feature.isChecked():
                        second_feature = track_group[self.second_feature_selected].to_numpy()
                    if self.cbs[2].currentText().startswith('t') and not self.abs_time_checkbox.isChecked():
                        t0 = math.floor(track_group[self.cbs[2].currentText()].to_numpy()[0])
                        timeline -= t0
                    elif self.cbs[2].currentText() == 'first detection' and not self.abs_time_checkbox.isChecked():

                        if 'area' in list(track_group.columns):
                            print('area in list')
                            feat = track_group['area'].values
                        else:
                            feat = feature

                        first_detection = timeline[feat == feat][0]
                        timeline -= first_detection
                        print(first_detection, timeline)

                    elif self.abs_time_checkbox.isChecked():
                        timeline -= int(self.frame_slider.value())

                    loc_t = [np.where(mapping == t)[0][0] for t in timeline]
                    matrix[cid, loc_t] = feature
                    if second_feature:
                        matrix[cid, loc_t + 1] = second_feature
                    print(timeline, loc_t)

                    cid += 1
                except:
                    pass
        return matrix

    def col_mean(self, matrix):

        mean_line = np.zeros(matrix.shape[1])
        mean_line[:] = np.nan
        std_line = np.copy(mean_line)

        for k in range(matrix.shape[1]):
            values = matrix[:, k]
            # values = values[values!=0]
            if len(values[values == values]) > 2:
                mean_line[k] = np.nanmean(values)
                std_line[k] = np.nanstd(values)

        return mean_line, std_line

    def initialize_axis(self):
        plt.rcParams['svg.fonttype'] = 'none'
        plt.rcParams["font.family"] = "sans-serif"
        SMALL_SIZE = 5
        MEDIUM_SIZE = 6
        BIGGER_SIZE = 7

        plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
        plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
        plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

        fig, ax = plt.subplots(1, 1, figsize=(10, 5))
        if self.well_option > 1:
            if self.check_class.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              hue=self.cbs[1].currentText(),
                              dodge=True, palette=self.palette)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  hue=self.cbs[1].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.palette)
            elif self.check_group.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              hue=self.cbs[2].currentText(),
                              dodge=True, palette=self.hue_colors)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  hue=self.cbs[2].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.hue_colors)
            else:
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              color=self.palette[0],
                              dodge=True)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  color=self.palette[0],
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3)
        else:
            if self.check_class.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              hue=self.cbs[1].currentText(),
                              dodge=True, palette=self.palette)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  hue=self.cbs[1].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.palette)
            elif self.check_group.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              hue=self.cbs[2].currentText(),
                              dodge=True, palette=self.hue_colors)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  hue=self.cbs[2].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.hue_colors)
            else:
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              color=self.palette[0],
                              dodge=True)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  color=self.palette[0],
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3)
        self.ax.spines['top'].set_visible(False)
        self.ax.spines['right'].set_visible(False)
        self.ax.set_ylabel(self.feature_selected)
        # ax.set_ylim(0.95,1.3)
        plt.tight_layout()

    def plot_survivals(self, id):
        self.ax.clear()
        plt.rcParams['svg.fonttype'] = 'none'
        plt.rcParams["font.family"] = "sans-serif"
        SMALL_SIZE = 5
        MEDIUM_SIZE = 6
        BIGGER_SIZE = 7

        plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
        plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
        plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

        fig, ax = plt.subplots(1, 1, figsize=(10, 5))
        if self.well_option > 1:
            if self.check_class.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              hue=self.cbs[1].currentText(),
                              dodge=True, palette=self.palette)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  hue=self.cbs[1].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.palette)
            elif self.check_group.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              hue=self.cbs[2].currentText(),
                              dodge=True, palette=self.hue_colors)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  hue=self.cbs[2].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.hue_colors)
            else:
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                              color=self.palette[0],
                              dodge=True)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='well_name',
                                  color=self.palette[0],
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3)
        else:
            if self.check_class.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              hue=self.cbs[1].currentText(),
                              dodge=True, palette=self.palette)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  hue=self.cbs[1].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.palette)
            elif self.check_group.isChecked():
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              hue=self.cbs[2].currentText(),
                              dodge=True, palette=self.hue_colors)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  hue=self.cbs[2].currentText(),
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3, palette=self.hue_colors)
            else:
                sns.boxenplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                              color=self.palette[0],
                              dodge=True)
                if self.show_cell_lines:
                    sns.stripplot(ax=self.ax, data=self.plot_data, y=self.feature_selected, x='position',
                                  color=self.palette[0],
                                  dodge=True, alpha=0.3, linewidth=0.5, size=3)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_ylabel(self.feature_selected)
        plt.tight_layout()
        self.survival_window.setMinimumHeight(int(0.5 * self.screen_height))
        self.survival_window.setMinimumWidth(int(0.8 * self.survivalWidget.width()))
        self.survival_window.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.survival_window.canvas.draw()

    def plot_line(self, line, color, label, mean_signal, ci_option=True, cell_lines_option=False, alpha_ci=0.5,
                  alpha_cell_lines=0.5, std_signal=None, matrix=None):
        try:
            if 'second' in str(mean_signal):
                self.ax2.plot(line['timeline'] * self.FrameToMin, line[mean_signal], color=color, label=label)
            else:
                self.ax.plot(line['timeline'] * self.FrameToMin, line[mean_signal], color=color, label=label)
            if ci_option and std_signal is not None:
                if 'second' in str(mean_signal):
                    self.ax2.fill_between(line['timeline'] * self.FrameToMin,
                                          [a - b for a, b in zip(line[mean_signal], line[std_signal])],
                                          [a + b for a, b in zip(line[mean_signal], line[std_signal])],
                                          color=color,
                                          alpha=alpha_ci,
                                          )
                else:
                    self.ax.fill_between(line['timeline'] * self.FrameToMin,
                                         [a - b for a, b in zip(line[mean_signal], line[std_signal])],
                                         [a + b for a, b in zip(line[mean_signal], line[std_signal])],
                                         color=color,
                                         alpha=alpha_ci,
                                         )
            if cell_lines_option and matrix is not None:
                print(mean_signal)
                mat = line[matrix]
                if 'second' in str(mean_signal):
                    for i in range(mat.shape[0]):
                        self.ax2.plot(line['timeline'] * self.FrameToMin, mat[i, :], color=color,
                                      alpha=alpha_cell_lines)
                else:
                    for i in range(mat.shape[0]):
                        self.ax.plot(line['timeline'] * self.FrameToMin, mat[i, :], color=color, alpha=alpha_cell_lines)

        except Exception as e:
            print(f'Exception {e}')

    def switch_to_log(self):

        """
        Switch threshold histogram to log scale. Auto adjust.
        """

        if self.ax.get_yscale() == 'linear':
            self.ax.set_yscale('log')
        # self.ax.set_ylim(0.01,1.05)
        else:
            self.ax.set_yscale('linear')
        # self.ax.set_ylim(0.01,1.05)

        # self.ax.autoscale()
        self.survival_window.canvas.draw_idle()

    def show_hide_legend(self):
        if self.legend_visible:
            self.ax.legend().set_visible(False)
            self.legend_visible = False
            self.legend_btn.setIcon(icon(MDI6.text_box_outline, color="black"))
        else:
            self.ax.legend().set_visible(True)
            self.legend_visible = True
            self.legend_btn.setIcon(icon(MDI6.text_box, color="black"))

        self.survival_window.canvas.draw_idle()

    def look_for_metadata(self):

        self.metadata_found = False
        self.metafiles = glob(self.exp_dir + os.sep.join([f'W*', '*', 'movie', '*metadata.txt'])) \
                         + glob(self.exp_dir + os.sep.join([f'W*', '*', '*metadata.txt'])) \
                         + glob(self.exp_dir + os.sep.join([f'W*', '*metadata.txt'])) \
                         + glob(self.exp_dir + '*metadata.txt')
        print(f'Found {len(self.metafiles)} metadata files...')
        if len(self.metafiles) > 0:
            self.metadata_found = True

    def switch_selection_mode(self, id):
        print(f'button {id} was clicked')
        for i in range(2):
            if self.select_option[i].isChecked():
                self.selection_mode = self.select_label[i]
        if self.selection_mode == 'name':
            if len(self.metafiles) > 0:
                self.position_scatter.hide()
            self.line_choice_widget.show()
        else:
            if len(self.metafiles) > 0:
                self.position_scatter.show()
            self.line_choice_widget.hide()

    def load_coordinates(self):

        """
        Read metadata and try to extract position coordinates
        """

        self.no_meta = False
        try:
            with open(self.metafiles[0], 'r') as f:
                data = json.load(f)
                positions = data['Summary']['InitialPositionList']
        except Exception as e:
            print(f'Trouble loading metadata: error {e}...')
            return None

        for k in range(len(positions)):
            pos_label = positions[k]['Label']
            try:
                coords = positions[k]['DeviceCoordinatesUm']['XYStage']
            except:
                try:
                    coords = positions[k]['DeviceCoordinatesUm']['PIXYStage']
                except:
                    self.no_meta = True

            if not self.no_meta:
                files = self.df_pos_info['stack_path'].values
                pos_loc = [pos_label in f for f in files]
                self.df_pos_info.loc[pos_loc, 'x'] = coords[0]
                self.df_pos_info.loc[pos_loc, 'y'] = coords[1]
                self.df_pos_info.loc[pos_loc, 'metadata_tag'] = pos_label

    def update_annot(self, ind):

        pos = self.sc.get_offsets()[ind["ind"][0]]
        self.annot.xy = pos
        text = self.scat_labels[ind["ind"][0]]
        self.annot.set_text(text)
        self.annot.get_bbox_patch().set_facecolor('k')
        self.annot.get_bbox_patch().set_alpha(0.4)

    def hover(self, event):
        vis = self.annot.get_visible()
        if event.inaxes == self.ax_scatter:
            cont, ind = self.sc.contains(event)
            if cont:
                self.update_annot(ind)
                self.annot.set_visible(True)
                self.fig_scatter.canvas.draw_idle()
            else:
                if vis:
                    self.annot.set_visible(False)
                    self.fig_scatter.canvas.draw_idle()

    def unselect_position(self, event):
        print('unselecting position')
        self.survival_window.canvas.clear()
        ind = event.ind  # index of selected position
        well_idx = self.df_pos_info.iloc[ind]['well_index'].values[0]
        selectedPos = self.df_pos_info.iloc[ind]['pos_path'].values[0]
        currentSelState = self.df_pos_info.iloc[ind]['select'].values[0]
        if self.plot_options[0].isChecked() or self.plot_options[2].isChecked():
            self.df_pos_info.loc[self.df_pos_info['well_index'] == well_idx, 'select'] = not currentSelState
            self.df_well_info.loc[self.df_well_info['well_index'] == well_idx, 'select'] = not currentSelState
            if len(self.well_indices) > 1:
                self.well_display_options[well_idx].setChecked(not currentSelState)
            else:
                for p in self.pos_display_options:
                    p.setChecked(not currentSelState)
        else:
            self.df_pos_info.loc[self.df_pos_info['pos_path'] == selectedPos, 'select'] = not currentSelState
            if len(self.well_indices) <= 1:
                self.pos_display_options[ind[0]].setChecked(not currentSelState)

        self.sc.set_color(self.select_color(self.df_pos_info["select"].values))
        self.position_scatter.canvas.draw_idle()
        self.plot_survivals(0)

    def select_survival_lines(self):
        if self.plot_mode == 'wells':
            selected_wells = []
            for i in range(len(self.well_display_options)):
                if self.well_display_options[i].isChecked():
                    selected_wells.append(i)
                self.df_well_info.loc[self.df_well_info['well_index'] == i, 'select'] = self.well_display_options[
                    i].isChecked()
                self.df_pos_info.loc[self.df_pos_info['well_index'] == i, 'select'] = self.well_display_options[
                    i].isChecked()
            if len(selected_wells) == 0:
                print('No wells selected')
                self.ax.clear()
            else:
                self.ax.clear()
                self.well_indices = selected_wells
                self.compute_signal_functions()
                self.plot_survivals(0)
        else:
            selected_pos = []
            for i in range(len(self.pos_display_options)):
                if self.pos_display_options[i].isChecked():
                    selected_pos.append(i + 1)
                self.df_pos_info.loc[self.df_pos_info['pos_index'] == i, 'select'] = self.pos_display_options[
                    i].isChecked()
            if len(selected_pos) == 0:
                self.ax.clear()
            else:
                self.position_indices = selected_pos
                self.ax.clear()
                self.compute_signal_functions()
                self.plot_survivals(0)

        if len(self.metafiles) > 0:
            self.sc.set_color(self.select_color(self.df_pos_info["select"].values))
            self.position_scatter.canvas.draw_idle()

    def select_color(self, selection):
        colors = [tab10(0) if s else tab10(0.1) for s in selection]
        return colors

    def plot_spatial_location(self):

        try:
            self.sc = self.ax_scatter.scatter(self.df_pos_info["x"].values, self.df_pos_info["y"].values, picker=True,
                                              pickradius=1, color=self.select_color(self.df_pos_info["select"].values))
            self.scat_labels = self.df_pos_info['metadata_tag'].values
            self.ax_scatter.invert_xaxis()
            self.annot = self.ax_scatter.annotate("", xy=(0, 0), xytext=(10, 10), textcoords="offset points",
                                                  bbox=dict(boxstyle="round", fc="w"),
                                                  arrowprops=dict(arrowstyle="->"))
            self.annot.set_visible(False)
            self.fig_scatter.canvas.mpl_connect("motion_notify_event", self.hover)
            self.fig_scatter.canvas.mpl_connect("pick_event", self.unselect_position)
        except Exception as e:
            pass

    def switch_ref_time_mode(self):
        if self.abs_time_checkbox.isChecked():
            self.frame_slider.setEnabled(True)
            self.cbs[-1].setEnabled(False)
        else:
            self.frame_slider.setEnabled(False)
            self.cbs[-1].setEnabled(True)

    def switch_ci(self):

        if self.show_ci:
            self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal, color="black"))
        else:
            self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal, color="blue"))
        self.show_ci = not self.show_ci
        self.plot_survivals(0)

    def switch_cell_lines(self):

        if self.show_cell_lines:
            self.cell_lines_btn.setIcon(icon(MDI6.view_headline, color="black"))
        else:
            self.cell_lines_btn.setIcon(icon(MDI6.view_headline, color="blue"))
        self.show_cell_lines = not self.show_cell_lines
        self.plot_survivals(0)

    def set_class_to_plot(self):

        if self.all_btn.isChecked():
            self.target_class = [0, 1]
        elif self.event_btn.isChecked():
            self.target_class = [0]
        else:
            self.target_class = [1]

        self.plot_survivals(0)



"""
Copright © 2023 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

from PyQt5.QtWidgets import QMainWindow, QComboBox, QLabel, QRadioButton, QLineEdit, QApplication, QPushButton, QScrollArea, QWidget, QVBoxLayout, QHBoxLayout
from PyQt5.QtCore import Qt, QSize
from celldetective.gui.gui_utils import center_window, QHSeperationLine
from superqt import QLabeledDoubleSlider, QLabeledSlider
from celldetective.utils import extract_experiment_channels, get_software_location
import json
import numpy as np
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
import os
from celldetective.gui import Styles

class ConfigSignalAnnotator(QMainWindow, Styles):
	
	"""
	UI to set normalization and animation parameters for the annotator tool. 

	"""

	def __init__(self, parent_window=None):
		
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Configure signal annotator")
		self.mode = self.parent_window.mode
		self.exp_dir = self.parent_window.exp_dir
		self.soft_path = get_software_location()
		
		if self.mode=="targets":
			self.instructions_path = self.parent_window.exp_dir + "configs/signal_annotator_config_targets.json"
		elif self.mode=="effectors":
			self.instructions_path = self.parent_window.exp_dir + "configs/signal_annotator_config_effectors.json"
		elif self.mode == "neighborhood":
			self.instructions_path = self.parent_window.exp_dir + "configs/signal_annotator_config_neighborhood.json"

		exp_config = self.exp_dir +"config.ini"
		#self.config_path = self.exp_dir + self.config_name
		self.channel_names, self.channels = extract_experiment_channels(exp_config)
		self.channel_names = np.array(self.channel_names)
		self.channels = np.array(self.channels)
		self.log_option = False

		self.screen_height = self.parent_window.parent_window.parent_window.screen_height
		center_window(self)

		self.setMinimumHeight(int(0.4*self.screen_height))
		self.setMaximumHeight(int(0.8*self.screen_height))
		self.populate_widget()
		#self.load_previous_measurement_instructions()


	def populate_widget(self):
		
		"""
		Create the widgets.
		
		"""

		self.scroll_area = QScrollArea(self)
		self.button_widget = QWidget()

		self.main_layout = QVBoxLayout()
		self.main_layout.setContentsMargins(30,30,30,30)

		sub_layout = QVBoxLayout()
		sub_layout.setContentsMargins(10,10,10,20)

		self.button_widget.setLayout(self.main_layout)
		sub_layout.setContentsMargins(30,30,30,30)

		sub_layout.addWidget(QLabel('Modality: '))
		
		# Create radio buttons
		option_layout = QHBoxLayout()
		self.gs_btn = QRadioButton('grayscale')
		self.gs_btn.setChecked(True)
		option_layout.addWidget(self.gs_btn, alignment=Qt.AlignCenter)

		self.rgb_btn = QRadioButton('RGB')
		option_layout.addWidget(self.rgb_btn, alignment=Qt.AlignCenter)
		sub_layout.addLayout(option_layout)

		btn_hbox = QHBoxLayout()

		self.percentile_btn = QPushButton()
		self.percentile_btn.setIcon(icon(MDI6.percent_circle_outline,color="black"))
		self.percentile_btn.setIconSize(QSize(20, 20))	
		self.percentile_btn.setStyleSheet(self.button_select_all)	
		self.percentile_btn.setToolTip("Switch to percentile normalization values.")
		self.percentile_btn.clicked.connect(self.switch_to_absolute_normalization_mode)

		self.log_btn = QPushButton()
		self.log_btn.setIcon(icon(MDI6.math_log,color="black"))
		self.log_btn.setStyleSheet(self.button_select_all)
		self.log_btn.clicked.connect(self.switch_to_log)
		self.log_btn.setToolTip("Log-transform the intensities.")
		self.log_btn.setIconSize(QSize(20, 20))	

		btn_hbox.addWidget(QLabel(''), 90)
		btn_hbox.addWidget(self.log_btn, 5,alignment=Qt.AlignRight)
		btn_hbox.addWidget(self.percentile_btn, 5,alignment=Qt.AlignRight)
		sub_layout.addLayout(btn_hbox)

		self.channel_cbs = [QComboBox() for i in range(3)]
		self.channel_cbs_lbls = [QLabel() for i in range(3)]

		self.min_val_les = [QLineEdit('0') for i in range(3)]
		self.min_val_lbls = [QLabel('Min value: ') for i in range(3)]
		self.max_val_les = [QLineEdit('10000') for i in range(3)]
		self.max_val_lbls = [QLabel('Max value: ') for i in range(3)]
		self.percentile_mode = False

		self.rgb_text = ['R: ', 'G: ', 'B: ']

		for i in range(3):

			hlayout = QHBoxLayout()
			self.channel_cbs[i].addItems(self.channel_names)
			self.channel_cbs[i].setCurrentIndex(i)
			self.channel_cbs_lbls[i].setText(self.rgb_text[i])
			hlayout.addWidget(self.channel_cbs_lbls[i], 20)
			hlayout.addWidget(self.channel_cbs[i], 80)
			sub_layout.addLayout(hlayout)

			hlayout2 = QHBoxLayout()
			hlayout2.addWidget(self.min_val_lbls[i], 20)
			hlayout2.addWidget(self.min_val_les[i], 80)			
			sub_layout.addLayout(hlayout2)

			hlayout3 = QHBoxLayout()
			hlayout3.addWidget(self.max_val_lbls[i], 20)
			hlayout3.addWidget(self.max_val_les[i], 80)			
			sub_layout.addLayout(hlayout3)

		self.enable_channels()
		
		self.gs_btn.toggled.connect(self.enable_channels)
		self.rgb_btn.toggled.connect(self.enable_channels)

		self.hsep = QHSeperationLine()
		sub_layout.addWidget(self.hsep)

		hbox_frac = QHBoxLayout()
		hbox_frac.addWidget(QLabel('fraction: '), 20)

		self.fraction_slider = QLabeledDoubleSlider()
		self.fraction_slider.setSingleStep(0.05)
		self.fraction_slider.setTickInterval(0.05)
		self.fraction_slider.setSingleStep(1)
		self.fraction_slider.setOrientation(1)
		self.fraction_slider.setRange(0.1,1)
		self.fraction_slider.setValue(0.25)

		hbox_frac.addWidget(self.fraction_slider, 80)
		sub_layout.addLayout(hbox_frac)


		hbox_interval = QHBoxLayout()
		hbox_interval.addWidget(QLabel('interval [ms]: '), 20)

		self.interval_slider = QLabeledSlider()
		self.interval_slider.setSingleStep(1)
		self.interval_slider.setTickInterval(1)
		self.interval_slider.setSingleStep(1)
		self.interval_slider.setOrientation(1)
		self.interval_slider.setRange(1,1000)
		self.interval_slider.setValue(1)
		hbox_interval.addWidget(self.interval_slider, 80)
		sub_layout.addLayout(hbox_interval)

		self.main_layout.addLayout(sub_layout)

		self.submit_btn = QPushButton('Save')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.write_instructions)
		self.main_layout.addWidget(self.submit_btn)


		self.button_widget.adjustSize()
		self.scroll_area.setAlignment(Qt.AlignCenter)
		self.scroll_area.setWidget(self.button_widget)
		self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
		self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
		self.scroll_area.setWidgetResizable(True)
		self.setCentralWidget(self.scroll_area)
		self.show()

		self.read_instructions()

		QApplication.processEvents()

	def enable_channels(self):

		"""
		Enable three channels when RGB mode is checked.
		
		"""

		if self.gs_btn.isChecked():

			self.log_btn.setEnabled(True)
			self.percentile_btn.setEnabled(False)

			for k in range(1,3):
				self.channel_cbs[k].setEnabled(False)
				self.channel_cbs_lbls[k].setEnabled(False)

			for k in range(3):

				self.min_val_les[k].setEnabled(False)
				self.min_val_lbls[k].setEnabled(False)
				self.max_val_les[k].setEnabled(False)
				self.max_val_lbls[k].setEnabled(False)

		elif self.rgb_btn.isChecked():

			self.log_btn.setEnabled(False)
			self.percentile_btn.setEnabled(True)

			for k in range(3):

				self.channel_cbs[k].setEnabled(True)
				self.channel_cbs_lbls[k].setEnabled(True)

				self.min_val_les[k].setEnabled(True)
				self.min_val_lbls[k].setEnabled(True)
				self.max_val_les[k].setEnabled(True)
				self.max_val_lbls[k].setEnabled(True)

	def switch_to_absolute_normalization_mode(self):

		"""
		Use absolute or percentile values for the normalization of each individual channel.
		
		"""

		if self.percentile_mode:
			self.percentile_mode = False
			self.percentile_btn.setIcon(icon(MDI6.percent_circle_outline,color="black"))
			self.percentile_btn.setIconSize(QSize(20, 20))		
			self.percentile_btn.setToolTip("Switch to percentile normalization values.")	
			for k in range(3):
				self.min_val_lbls[k].setText('Min value: ')	
				self.min_val_les[k].setText('0')	
				self.max_val_lbls[k].setText('Max value: ')	
				self.max_val_les[k].setText('10000')
		else:
			self.percentile_mode = True
			self.percentile_btn.setIcon(icon(MDI6.percent_circle,color="black"))
			self.percentile_btn.setIconSize(QSize(20, 20))
			self.percentile_btn.setToolTip("Switch to absolute normalization values.")	
			for k in range(3):
				self.min_val_lbls[k].setText('Min percentile: ')
				self.min_val_les[k].setText('0.01')	
				self.max_val_lbls[k].setText('Max percentile: ')	
				self.max_val_les[k].setText('99.99')

	def write_instructions(self):

		"""
		Save the current configuration.
		
		"""

		instructions = {'rgb_mode': self.rgb_btn.isChecked(), 'percentile_mode': self.percentile_mode, 'fraction': float(self.fraction_slider.value()), 'interval': int(self.interval_slider.value()), 'log': self.log_option}
		max_i = 3 if self.rgb_btn.isChecked() else 1
		channels = []
		for i in range(max_i):
			channels.append([self.channel_cbs[i].currentText(), float(self.min_val_les[i].text()), float(self.max_val_les[i].text())])
		instructions.update({'channels': channels})
		
		print('Instructions: ', instructions)
		file_name = self.instructions_path
		with open(file_name, 'w') as f:
			json.dump(instructions, f, indent=4)
		print('Done.')
		self.close()

	def read_instructions(self):
		
		"""
		Read and set the widgets to the last configuration.

		"""

		print('Reading instructions..')
		if os.path.exists(self.instructions_path):
			with open(self.instructions_path, 'r') as f:
				instructions = json.load(f)
				print(instructions)
				
				if 'rgb_mode' in instructions:
					rgb_mode = instructions['rgb_mode']
					if rgb_mode:
						self.rgb_btn.setChecked(True)
						self.gs_btn.setChecked(False)

				if 'percentile_mode' in instructions:
					percentile_mode = instructions['percentile_mode']
					if percentile_mode:
						self.percentile_btn.click()

				if 'channels' in instructions:
					channels = instructions['channels']
					
					if len(channels)==1:
						max_iter = 1
					else:
						max_iter = 3

					for i in range(max_iter):
						idx = self.channel_cbs[i].findText(channels[i][0])
						self.channel_cbs[i].setCurrentIndex(idx)					
						self.min_val_les[i].setText(str(channels[i][1]))
						self.max_val_les[i].setText(str(channels[i][2]))

				if 'fraction' in instructions:
					fraction = instructions['fraction']
					self.fraction_slider.setValue(fraction)

				if 'interval' in instructions:
					interval = instructions['interval']
					self.interval_slider.setValue(interval)

				if 'log' in instructions:
					self.log_option = not instructions['log']
					self.switch_to_log()

	def switch_to_log(self):

		"""
		Switch threshold histogram to log scale. Auto adjust.
		"""

		if not self.log_option:
			self.log_btn.setIcon(icon(MDI6.math_log,color="#1565c0"))
			self.log_option = True
		else:
			self.log_btn.setIcon(icon(MDI6.math_log,color="black"))
			self.log_option = False


"""
Copright © 2023 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

import argparse
import os
import shutil
from glob import glob
import json
from tqdm import tqdm
import numpy as np
import random

from celldetective.utils import load_image_dataset, augmenter, interpolate_nan
from celldetective.io import normalize_multichannel
from stardist import fill_label_holes
from art import tprint
from distutils.dir_util import copy_tree
from csbdeep.utils import save_json

tprint("Train")

parser = argparse.ArgumentParser(description="Train a signal model from instructions.",
								formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-c',"--config", required=True,help="Training instructions")
parser.add_argument('-g',"--use_gpu", required=True, help="Use GPU")

args = parser.parse_args()
process_arguments = vars(args)
instructions = str(process_arguments['config'])
use_gpu = bool(process_arguments['use_gpu'])

if os.path.exists(instructions):
	with open(instructions, 'r') as f:
		training_instructions = json.load(f)
else:
	print('Training instructions could not be found. Abort.')
	os.abort()

model_name = training_instructions['model_name']
target_directory = training_instructions['target_directory']
model_type = training_instructions['model_type']
pretrained = training_instructions['pretrained']

datasets = training_instructions['ds']

target_channels = training_instructions['channel_option']
normalization_percentile = training_instructions['normalization_percentile']
normalization_clip = training_instructions['normalization_clip']
normalization_values = training_instructions['normalization_values']
spatial_calibration = training_instructions['spatial_calibration']

validation_split = training_instructions['validation_split']
augmentation_factor = training_instructions['augmentation_factor']

learning_rate = training_instructions['learning_rate']
epochs = training_instructions['epochs']
batch_size = training_instructions['batch_size']


# Load dataset
print(f'Datasets: {datasets}')
X,Y,filenames = load_image_dataset(datasets, target_channels, train_spatial_calibration=spatial_calibration,
						mask_suffix='labelled')
print('Dataset loaded...')

values = []
percentiles = []
for k in range(len(normalization_percentile)):
	if normalization_percentile[k]:
		percentiles.append(normalization_values[k])
		values.append(None)
	else:
		percentiles.append(None)
		values.append(normalization_values[k])

X = [normalize_multichannel(x, **{"percentiles": percentiles, 'values': values, 'clip': normalization_clip}) for x in X]

for k in range(len(X)):
	x = X[k].copy()
	x_interp = np.moveaxis([interpolate_nan(x[:,:,c].copy()) for c in range(x.shape[-1])],0,-1)
	X[k] = x_interp

Y = [fill_label_holes(y) for y in tqdm(Y)]

assert len(X) > 1, "not enough training data"
rng = np.random.RandomState()
ind = rng.permutation(len(X))
n_val = max(1, int(round(validation_split * len(ind))))
ind_train, ind_val = ind[:-n_val], ind[-n_val:]
X_val, Y_val = [X[i] for i in ind_val]  , [Y[i] for i in ind_val]
X_trn, Y_trn = [X[i] for i in ind_train], [Y[i] for i in ind_train]

files_train = [filenames[i] for i in ind_train]
files_val = [filenames[i] for i in ind_val]

print('number of images: %3d' % len(X))
print('- training:       %3d' % len(X_trn))
print('- validation:     %3d' % len(X_val))

if model_type=='cellpose':

	# do augmentation in place
	X_aug = []; Y_aug = [];
	n_val = max(1, int(round(augmentation_factor * len(X_trn))))
	indices = random.choices(list(np.arange(len(X_trn))), k=n_val)
	print('Performing image augmentation pre-training...')
	for i in tqdm(indices):
		x_aug,y_aug = augmenter(X_trn[i], Y_trn[i])
		X_aug.append(x_aug)
		Y_aug.append(y_aug)
	
	# Channel axis in front for cellpose
	X_aug = [np.moveaxis(x,-1,0) for x in X_aug]
	X_val = [np.moveaxis(x,-1,0) for x in X_val]
	print('number of augmented images: %3d' % len(X_aug))

	from cellpose.models import CellposeModel
	from cellpose.io import logger_setup
	import torch
	
	if not use_gpu:
		print('Using CPU for training...')
		device = torch.device("cpu")
	else:
		print('Using GPU for training...')
		
	logger, log_file = logger_setup()
	print(f'Pretrained model: ',pretrained)
	if pretrained is not None:
		pretrained_path = os.sep.join([pretrained,os.path.split(pretrained)[-1]])
	else:
		pretrained_path = pretrained
	
	model = CellposeModel(gpu=use_gpu, model_type=None, pretrained_model=pretrained_path, diam_mean=30.0, nchan=X_aug[0].shape[0],) 
	model.train(train_data=X_aug, train_labels=Y_aug, normalize=False, channels=None, batch_size=batch_size,
				min_train_masks=1,save_path=target_directory+os.sep+model_name,n_epochs=epochs, model_name=model_name, learning_rate=learning_rate, test_data = X_val, test_labels=Y_val)

	file_to_move = glob(os.sep.join([target_directory, model_name, 'models','*']))[0]
	shutil.move(file_to_move, os.sep.join([target_directory, model_name,''])+os.path.split(file_to_move)[-1])
	os.rmdir(os.sep.join([target_directory, model_name, 'models']))

	diameter = model.diam_labels

	if pretrained is not None and os.path.split(pretrained)[-1]=='CP_nuclei':
		standard_diameter = 17.0
	else:
		standard_diameter = 30.0

	input_spatial_calibration = spatial_calibration #*diameter / standard_diameter

	config_inputs = {"channels": target_channels, "diameter": standard_diameter, 'cellprob_threshold': 0., 'flow_threshold': 0.4,
	'normalization_percentile': normalization_percentile, 'normalization_clip': normalization_clip,
	'normalization_values': normalization_values, 'model_type': 'cellpose',
	'spatial_calibration': input_spatial_calibration, 'dataset': {'train': files_train, 'validation': files_val}}
	json_input_config = json.dumps(config_inputs, indent=4)
	with open(os.sep.join([target_directory, model_name, "config_input.json"]), "w") as outfile:
		outfile.write(json_input_config)

elif model_type=='stardist':
	
	from stardist import calculate_extents, gputools_available
	from stardist.models import Config2D, StarDist2D
	
	n_rays = 32
	print(gputools_available())
	
	n_channel=X_trn[0].shape[-1]
	
	# Predict on subsampled grid for increased efficiency and larger field of view
	grid = (2,2)
	conf = Config2D(
		n_rays       = n_rays,
		grid         = grid,
		use_gpu      = use_gpu,
		n_channel_in = n_channel,
		train_learning_rate = learning_rate,
		train_patch_size = (256,256),
		train_epochs = epochs,
		train_reduce_lr = {'factor': 0.1, 'patience': 30, 'min_delta': 0},
		train_batch_size = batch_size,
		train_steps_per_epoch = int(augmentation_factor*len(X_trn)),
	)
	
	if use_gpu:
		from csbdeep.utils.tf import limit_gpu_memory
		limit_gpu_memory(None, allow_growth=True)

	if pretrained is None:
		model = StarDist2D(conf, name=model_name, basedir=target_directory)
	else:

		os.rename(instructions, os.sep.join([target_directory, model_name, 'temp.json']))
		copy_tree(pretrained, os.sep.join([target_directory, model_name]))
		
		if os.path.exists(os.sep.join([target_directory, model_name, 'training_instructions.json'])):
			os.remove(os.sep.join([target_directory, model_name, 'training_instructions.json']))
		if os.path.exists(os.sep.join([target_directory, model_name, 'config_input.json'])):
			os.remove(os.sep.join([target_directory, model_name, 'config_input.json']))
		if os.path.exists(os.sep.join([target_directory, model_name, 'logs'+os.sep])):
			shutil.rmtree(os.sep.join([target_directory, model_name, 'logs']))
		os.rename(os.sep.join([target_directory, model_name, 'temp.json']),os.sep.join([target_directory, model_name, 'training_instructions.json']))

		#shutil.copytree(pretrained, os.sep.join([target_directory, model_name]))
		model = StarDist2D(None, name=model_name, basedir=target_directory)
		model.config.train_epochs = epochs
		model.config.train_batch_size = min(len(X_trn),batch_size)
		model.config.train_learning_rate = learning_rate # perf seems bad if lr is changed in transfer
		model.config.use_gpu = use_gpu
		model.config.train_reduce_lr = {'factor': 0.1, 'patience': 10, 'min_delta': 0}
		print(f'{model.config=}')

		save_json(vars(model.config), os.sep.join([target_directory, model_name, 'config.json']))

	median_size = calculate_extents(list(Y_trn), np.mean)
	fov = np.array(model._axes_tile_overlap('YX'))
	print(f"median object size:      {median_size}")
	print(f"network field of view :  {fov}")
	if any(median_size > fov):
		print("WARNING: median object size larger than field of view of the neural network.")

	if augmentation_factor==1.0:
		model.train(X_trn, Y_trn, validation_data=(X_val,Y_val))
	else:
		model.train(X_trn, Y_trn, validation_data=(X_val,Y_val), augmenter=augmenter)
	model.optimize_thresholds(X_val,Y_val)

	config_inputs = {"channels": target_channels, 'normalization_percentile': normalization_percentile,
	'normalization_clip': normalization_clip, 'normalization_values': normalization_values, 
	'model_type': 'stardist', 'spatial_calibration': spatial_calibration, 'dataset': {'train': files_train, 'validation': files_val}}

	json_input_config = json.dumps(config_inputs, indent=4)
	with open(os.sep.join([target_directory, model_name, "config_input.json"]), "w") as outfile:
		outfile.write(json_input_config)

print('Done.')







"""
Copyright © 2022 Laboratoire Adhesion et Inflammation, Authored by Remy Torro, 
Ksenija Dervanova.

Add extra properties to measure using regionprops (measure_features). 
The functions must take "regionmask" as the first argument, corresponding
to the binary mask of each cell of interest generated by regionprops 
and an optional "intensity_image" second argument, i.e. the associated 
multichannel image crop. The measurement will be named after the function
name. If a function function() has several outputs (e.g. due to a multichannel image),
they will be labelled with an index as function-0, function-1... A routine 
automatically replacements keywords "intensity-0", "intensity-1" and so on,
with the actual name of the intensity channel (e.g. "intensity-0" -> 
"brightfield_channel"). Due to this indexing behavior, avoid using digits 
in the function names and prefer text instead: "intensity-99()" --> 
"intensity-ninety-nine()".

New functions appear automatically in the ConfigMeasurements widget of the
GUI.

"""

import warnings

import numpy as np
from scipy.ndimage import distance_transform_edt, center_of_mass
from scipy.spatial.distance import euclidean
from celldetective.utils import interpolate_nan, contour_of_instance_segmentation

# Percentiles

def intensity_percentile_ninety_nine(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],99)

def intensity_percentile_ninety_five(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],95)

def intensity_percentile_ninety(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],90)

def intensity_percentile_seventy_five(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],75)

def intensity_percentile_fifty(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],50)

def intensity_percentile_twenty_five(regionmask, intensity_image):
	return np.nanpercentile(intensity_image[regionmask],25)

# STD

def intensity_std(regionmask, intensity_image):
	return np.nanstd(intensity_image[regionmask])


def intensity_median(regionmask, intensity_image):
	return np.nanmedian(intensity_image[regionmask])

def intensity_nanmean(regionmask, intensity_image):
	return np.nanmean(intensity_image[regionmask])

def intensity_centre_of_mass_displacement(regionmask, intensity_image):

	intensity_image = interpolate_nan(intensity_image.copy())

	y, x = np.mgrid[:regionmask.shape[0], :regionmask.shape[1]]
	xtemp = x.copy()
	ytemp = y.copy()
	intensity_weighted_center = center_of_mass(intensity_image, regionmask)
	centroid_x = intensity_weighted_center[1]
	centroid_y = intensity_weighted_center[0]

	#centroid_x = np.sum(xtemp * intensity_image) / np.sum(intensity_image)
	geometric_centroid_x = np.sum(xtemp * regionmask) / np.sum(regionmask)
	geometric_centroid_y = np.sum(ytemp * regionmask) / np.sum(regionmask)
	try:
		distance = euclidean(np.array((geometric_centroid_y, geometric_centroid_x)), np.array((centroid_y, centroid_x)))
	except:
		distance = np.nan

	delta_x = geometric_centroid_x - centroid_x
	delta_y = geometric_centroid_y - centroid_y
	direction_arctan = np.arctan2(delta_y, delta_x) * 180 / np.pi
	if direction_arctan < 0:
		direction_arctan += 360

	return distance, direction_arctan, centroid_x - geometric_centroid_x, centroid_y - geometric_centroid_y

def intensity_radial_gradient(regionmask, intensity_image):

	try:
		warnings.filterwarnings('ignore', message="Polyfit may be poorly conditioned")
		cell_mask = regionmask.copy()
		intensity = intensity_image.copy()
		y = intensity[cell_mask].flatten()
		x = distance_transform_edt(cell_mask)
		x = x[cell_mask].flatten()
		params = np.polyfit(x, y, 1)
		line = np.poly1d(params)

		return line.coefficients[0], line.coefficients[1]
	except Exception as e:
		print(e)
		return np.nan, np.nan


def intensity_centre_of_mass_displacement_edge(regionmask, intensity_image):
	
	intensity_image = interpolate_nan(intensity_image.copy())
	edge_mask = contour_of_instance_segmentation(regionmask, 3)

	if np.sum(edge_mask)>0:
		
		y, x = np.mgrid[:edge_mask.shape[0], :edge_mask.shape[1]]
		xtemp = x.copy()
		ytemp = y.copy()
		intensity_weighted_center = center_of_mass(intensity_image, edge_mask)
		centroid_x = intensity_weighted_center[1]
		centroid_y = intensity_weighted_center[0]

		#centroid_x = np.sum(xtemp * intensity_image) / np.sum(intensity_image)
		geometric_centroid_x = np.sum(xtemp * regionmask) / np.sum(regionmask)
		geometric_centroid_y = np.sum(ytemp * regionmask) / np.sum(regionmask)
		
		try:
			distance = euclidean(np.array((geometric_centroid_y, geometric_centroid_x)), np.array((centroid_y, centroid_x)))
		except:
			distance = np.nan

		delta_x = geometric_centroid_x - centroid_x
		delta_y = geometric_centroid_y - centroid_y
		direction_arctan = np.arctan2(delta_y, delta_x) * 180 / np.pi
		if direction_arctan < 0:
			direction_arctan += 360

		return distance, direction_arctan, centroid_x - geometric_centroid_x, centroid_y - geometric_centroid_y
	else:
		return np.nan, np.nan, np.nan, np.nan


import numpy as np
from lifelines import KaplanMeierFitter

def switch_to_events(classes, event_times, max_times, origin_times=None, left_censored=True, FrameToMin=None):
	

	"""
	Converts time-to-event data into a format suitable for survival analysis, optionally adjusting for left censorship 
	and converting time units.

	This function processes event data by classifying each event based on whether it occurred or was censored by the end 
	of the observation period. It calculates the survival time for each event, taking into account the possibility of left 
	censorship and the option to convert time units (e.g., from frames to minutes).

	Parameters
	----------
	classes : array_like
		An array indicating the class of each event (e.g., 0 for event, 1 for non-event, 2 for else).
	event_times : array_like
		An array of times at which events occurred. For non-events, this might represent the time of last observation.
	max_times : array_like
		An array of maximum observation times for each event.
	origin_times : array_like, optional
		An array of origin times for each event. If None, origin times are assumed to be zero, and `left_censored` is 
		automatically set to False (default is None).
	left_censored : bool, optional
		Indicates whether to adjust for left censorship. If True, events with origin times are considered left-censored 
		if the origin time is zero (default is True).
	FrameToMin : float, optional
		A conversion factor to transform survival times from frames (or any other unit) to minutes. If None, no conversion 
		is applied (default is None).

	Returns
	-------
	tuple of lists
		A tuple containing two lists: `events` and `survival_times`. `events` is a list of binary indicators (1 for event 
		occurrence, 0 for censorship), and `survival_times` is a list of survival times corresponding to each event or 
		censorship.

	Notes
	-----
	- The function assumes that `classes`, `event_times`, `max_times`, and `origin_times` (if provided) are all arrays of 
	  the same length.
	- This function is particularly useful in preparing time-to-event data for survival analysis models, especially when 
	  dealing with censored data and needing to adjust time units.

	Examples
	--------
	>>> classes = [0, 1, 0]
	>>> event_times = [5, 10, 15]
	>>> max_times = [20, 20, 20]
	>>> origin_times = [0, 0, 5]
	>>> events, survival_times = switch_to_events_v2(classes, event_times, max_times, origin_times, FrameToMin=0.5)
	# This would process the events considering left censorship and convert survival times to minutes.
	
	"""

	events = []
	survival_times = []

	if origin_times is None:
		# then origin is zero
		origin_times = np.zeros_like(max_times)
		left_censored = False
		
	for c,t,mt,ot in zip(classes, event_times, max_times, origin_times):

		if left_censored:

			if ot>=0. and ot==ot:
				# origin time is larger than zero, no censorship
				if c==0 and t>0:
					delta_t = t - ot
					if delta_t>0:
						events.append(1)
						survival_times.append(delta_t)
					else:
						# negative delta t, invalid cell
						pass
				elif c==1:
					delta_t = mt - ot
					if delta_t>0:
						events.append(0)
						survival_times.append(delta_t)
					else:
						# negative delta t, invalid cell
						pass
				else:
					pass
			else:
				# origin time is zero, the event is left censored (we did not observe it start)
				pass

		else:
			if c==0 and t>0:
				events.append(1)
				survival_times.append(t - ot)
			elif c==1:
				events.append(0)
				survival_times.append(mt - ot)
			else:
				pass

	if FrameToMin is not None:
		#print('convert to minutes!', FrameToMin)
		survival_times = [s*FrameToMin for s in survival_times]
	return events, survival_times

def compute_survival(df, class_of_interest, t_event, t_reference=None, FrameToMin=1):

	cols = list(df.columns)
	assert class_of_interest in cols,"The requested class cannot be found in the dataframe..."
	assert t_event in cols,"The event time cannot be found in the dataframe..."
	left_censored = False

	classes = df.groupby(['position','TRACK_ID'])[class_of_interest].min().values
	event_times = df.groupby(['position','TRACK_ID'])[t_event].min().values
	max_times = df.groupby(['position','TRACK_ID'])['FRAME'].max().values

	if t_reference is not None:
		left_censored = True
		assert t_reference in cols,"The reference time cannot be found in the dataframe..."
		first_detections = df.groupby(['position','TRACK_ID'])[t_reference].max().values
	
	events, survival_times = switch_to_events(classes, event_times, max_times, origin_times=first_detections, left_censored=left_censored, FrameToMin=FrameToMin)
	ks = KaplanMeierFitter()
	if len(events)>0:
		ks.fit(survival_times, event_observed=events)
	else:
		ks = None

	return ks




import numpy as np
import pandas as pd
from tqdm import tqdm
from skimage.graph import pixel_graph
import os
from celldetective.utils import contour_of_instance_segmentation, extract_identity_col
from scipy.spatial.distance import cdist
from celldetective.io import locate_labels, get_position_pickle, get_position_table

abs_path = os.sep.join([os.path.split(os.path.dirname(os.path.realpath(__file__)))[0], 'celldetective'])


def set_live_status(setA, setB, status, not_status_option):
	"""
	Updates the live status for cells in two datasets based on specified status columns and options.

	This function assigns a live status to cells in two datasets (setA and setB) based on the provided
	status columns and options. If no status column is provided, all cells are marked as live. Otherwise,
	the function updates the datasets based on the status criteria, potentially inverting the status
	based on the `not_status_option`.

	Parameters
	----------
	setA : pandas.DataFrame
		The first dataset containing trajectory or position information for cells.
	setB : pandas.DataFrame
		The second dataset containing trajectory or position information for cells.
	status : list or None
		A list containing the names of the columns in setA and setB that classify cells as alive (1) or dead (0).
		If None, all cells are considered alive. The list should contain exactly two elements.
	not_status_option : list
		A list containing boolean values indicating whether to invert the status for setA and setB, respectively.
		True means the status should be inverted; False means it should not.

	Returns
	-------
	tuple
		A tuple containing the updated setA and setB DataFrames, along with the final status column names
		used to classify cells in each set.

	"""

	print(f"Provided statuses: {status}...")
	if status is None or status==["live_status","live_status"] or status==[None,None]:
		setA.loc[:,'live_status'] = 1
		setB.loc[:,'live_status'] = 1
		status = ['live_status', 'live_status']
	elif isinstance(status,list):
		assert len(status)==2,'Please provide only two columns to classify cells as alive or dead.'
		if status[0] is None or status[0]=='live_status':
			setA.loc[:,'live_status'] = 1
			status[0] = 'live_status'
		elif status[0] is not None and isinstance(not_status_option, list):
			setA.loc[setA[status[0]] == 2, status[0]] = 1  # already happened events become event
			if not_status_option[0]:
				setA.loc[:,'not_'+status[0]] = [not a if a==0 or a==1 else np.nan for a in setA.loc[:,status[0]].values]
				status[0] = 'not_'+status[0]
		if status[1] is None or status[1]=='live_status':
			setB.loc[:,'live_status'] = 1
			status[1] = 'live_status'
		elif status[1] is not None and isinstance(not_status_option, list):
			setB.loc[setB[status[1]] == 2, status[1]] = 1  # already happened events become event
			if not_status_option[1]:
				setB.loc[:, 'not_' + status[1]] = [not a if a == 0 or a == 1 else np.nan for a in
												   setB.loc[:, status[1]].values]
				status[1] = 'not_' + status[1]

		assert status[0] in list(setA.columns)
		assert status[1] in list(setB.columns)

	setA = setA.reset_index(drop=True)
	setB = setB.reset_index(drop=True)

	return setA, setB, status


def compute_attention_weight(dist_matrix, cut_distance, opposite_cell_status, opposite_cell_ids, axis=1,
							 include_dead_weight=True):
	"""
	Computes the attention weight for each cell based on its proximity to cells of an opposite type within a specified distance.

	This function calculates the attention weight for cells by considering the distance to the cells of an opposite type
	within a given cutoff distance. It optionally considers only the 'live' opposite cells based on their status. The function
	returns two arrays: one containing the attention weights and another containing the IDs of the closest opposite cells.

	Parameters
	----------
	dist_matrix : ndarray
		A 2D array representing the distance matrix between cells of two types.
	cut_distance : float
		The cutoff distance within which opposite cells will influence the attention weight.
	opposite_cell_status : ndarray
		An array indicating the status (e.g., live or dead) of each opposite cell. Only used when `include_dead_weight` is False.
	opposite_cell_ids : ndarray
		An array containing the IDs of the opposite cells.
	axis : int, optional
		The axis along which to compute the weights (default is 1). Axis 0 corresponds to rows, and axis 1 corresponds to columns.
	include_dead_weight : bool, optional
		If True, includes all opposite cells within the cutoff distance in the weight calculation, regardless of their status.
		If False, only considers opposite cells that are 'live' (default is True).

	Returns
	-------
	tuple of ndarrays
		A tuple containing two arrays: `weights` and `closest_opposite`. `weights` is an array of attention weights for each cell,
		and `closest_opposite` is an array of the IDs of the closest opposite cells within the cutoff distance.

	"""

	weights = np.empty(dist_matrix.shape[axis])
	closest_opposite = np.empty(dist_matrix.shape[axis])

	for i in range(dist_matrix.shape[axis]):
		if axis == 1:
			row = dist_matrix[:, i]
		elif axis == 0:
			row = dist_matrix[i, :]
		row[row == 0.] = 1.0E06
		nbr_opposite = len(row[row <= cut_distance])

		if not include_dead_weight:
			stat = opposite_cell_status[np.where(row <= cut_distance)[0]]
			nbr_opposite = len(stat[stat == 1])
			index_subpop = np.argmin(row[opposite_cell_status == 1])
			closest_opposite[i] = opposite_cell_ids[opposite_cell_status == 1][index_subpop]
		else:
			closest_opposite[i] = opposite_cell_ids[np.argmin(row)]

		if nbr_opposite > 0:
			weight = 1. / float(nbr_opposite)
			weights[i] = weight

	return weights, closest_opposite


def distance_cut_neighborhood(setA, setB, distance, mode='two-pop', status=None, not_status_option=None,
							  compute_cum_sum=True,
							  attention_weight=True, symmetrize=True, include_dead_weight=True,
							  column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X',
											 'y': 'POSITION_Y'}):
	"""

	Match neighbors in set A and B within a circle of radius d.

	Parameters
	----------
	setA,setB : pandas DataFrame
		Trajectory or position sets A and B.
	distance : float
		Cut-distance in pixels to match neighboring pairs.
	mode: str
		neighboring mode, between 'two-pop' (e.g. target-effector) and 'self' (target-target or effector-effector).
	status: None or status
		name to look for cells to ignore (because they are dead). By default all cells are kept.
	compute_cum_sum: bool,
		compute cumulated time of presence of neighbours (only if trajectories available for both sets)
	attention_weight: bool,
		compute the attention weight (how much a cell of set B is shared across cells of set A)
	symmetrize: bool,
		write in set B the neighborhood of set A
	include_dead_weight: bool
		do not count dead cells when establishing attention weight
	"""

	# Check live_status option
	if setA is not None and setB is not None:
		setA, setB, status = set_live_status(setA, setB, status, not_status_option)
	else:
		return None, None

	# Check distance option
	if not isinstance(distance, list):
		distance = [distance]

	for d in distance:
		# loop over each provided distance

		if mode == 'two-pop':
			neigh_col = f'neighborhood_2_circle_{d}_px'
		elif mode == 'self':
			neigh_col = f'neighborhood_self_circle_{d}_px'

		cl = []
		for s in [setA, setB]:

			# Check whether data can be tracked
			temp_column_labels = column_labels.copy()

			id_col = extract_identity_col(s)
			temp_column_labels.update({'track': id_col})
			if id_col=='ID':
				compute_cum_sum = False  # if no tracking data then cum_sum is not relevant
			cl.append(temp_column_labels)

			# Remove nan tracks (cells that do not belong to a track)
			s[neigh_col] = np.nan
			s[neigh_col] = s[neigh_col].astype(object)
			s.dropna(subset=[cl[-1]['track']], inplace=True)

		# Loop over each available timestep
		timeline = np.unique(np.concatenate([setA[cl[0]['time']].to_numpy(), setB[cl[1]['time']].to_numpy()])).astype(
			int)
		for t in tqdm(timeline):

			index_A = list(setA.loc[setA[cl[0]['time']] == t].index)
			coordinates_A = setA.loc[setA[cl[0]['time']] == t, [cl[0]['x'], cl[0]['y']]].to_numpy()
			ids_A = setA.loc[setA[cl[0]['time']] == t, cl[0]['track']].to_numpy()
			status_A = setA.loc[setA[cl[0]['time']] == t, status[0]].to_numpy()

			index_B = list(setB.loc[setB[cl[1]['time']] == t].index)
			coordinates_B = setB.loc[setB[cl[1]['time']] == t, [cl[1]['x'], cl[1]['y']]].to_numpy()
			ids_B = setB.loc[setB[cl[1]['time']] == t, cl[1]['track']].to_numpy()
			status_B = setB.loc[setB[cl[1]['time']] == t, status[1]].to_numpy()

			if len(ids_A) > 0 and len(ids_B) > 0:

				# compute distance matrix
				dist_map = cdist(coordinates_A, coordinates_B, metric="euclidean")

				if attention_weight:
					weights, closest_A = compute_attention_weight(dist_map, d, status_A, ids_A, axis=1,
																  include_dead_weight=include_dead_weight)

				# Target centric
				for k in range(dist_map.shape[0]):

					col = dist_map[k, :]
					col[col == 0.] = 1.0E06

					neighs_B = np.array([ids_B[i] for i in np.where((col <= d))[0]])
					status_neigh_B = np.array([status_B[i] for i in np.where((col <= d))[0]])
					dist_B = [round(col[i], 2) for i in np.where((col <= d))[0]]
					if len(dist_B) > 0:
						closest_B_cell = neighs_B[np.argmin(dist_B)]

					if symmetrize and attention_weight:
						n_neighs = float(len(neighs_B))
						if not include_dead_weight:
							n_neighs_alive = len(np.where(status_neigh_B == 1)[0])
							neigh_count = n_neighs_alive
						else:
							neigh_count = n_neighs
						if neigh_count > 0:
							weight_A = 1. / neigh_count
						else:
							weight_A = np.nan

						if not include_dead_weight and status_A[k] == 0:
							weight_A = 0

					neighs = []
					setA.at[index_A[k], neigh_col] = []
					for n in range(len(neighs_B)):

						# index in setB
						n_index = np.where(ids_B == neighs_B[n])[0][0]
						# Assess if neigh B is closest to A
						if attention_weight:
							if closest_A[n_index] == ids_A[k]:
								closest = True
							else:
								closest = False

						if symmetrize:
							# Load neighborhood previous data
							sym_neigh = setB.loc[index_B[n_index], neigh_col]
							if neighs_B[n] == closest_B_cell:
								closest_b = True
							else:
								closest_b = False
							if isinstance(sym_neigh, list):
								sym_neigh.append({'id': ids_A[k], 'distance': dist_B[n], 'status': status_A[k]})
							else:
								sym_neigh = [{'id': ids_A[k], 'distance': dist_B[n], 'status': status_A[k]}]
							if attention_weight:
								sym_neigh[-1].update({'weight': weight_A, 'closest': closest_b})

						# Write the minimum info about neighborhing cell B
						neigh_dico = {'id': neighs_B[n], 'distance': dist_B[n], 'status': status_neigh_B[n]}
						if attention_weight:
							neigh_dico.update({'weight': weights[n_index], 'closest': closest})

						if compute_cum_sum:
							# Compute the integrated presence of the neighboring cell B
							assert cl[1][
									   'track'] == 'TRACK_ID', 'The set B does not seem to contain tracked data. The cumulative time will be meaningless.'
							past_neighs = [[ll['id'] for ll in l] if len(l) > 0 else [None] for l in setA.loc[
								(setA[cl[0]['track']] == ids_A[k]) & (setA[cl[0]['time']] <= t), neigh_col].to_numpy()]
							past_neighs = [item for sublist in past_neighs for item in sublist]

							if attention_weight:
								past_weights = [[ll['weight'] for ll in l] if len(l) > 0 else [None] for l in setA.loc[
									(setA[cl[0]['track']] == ids_A[k]) & (
												setA[cl[0]['time']] <= t), neigh_col].to_numpy()]
								past_weights = [item for sublist in past_weights for item in sublist]

							cum_sum = len(np.where(past_neighs == neighs_B[n])[0])
							neigh_dico.update({'cumulated_presence': cum_sum + 1})

							if attention_weight:
								cum_sum_weighted = np.sum(
									[w if l == neighs_B[n] else 0 for l, w in zip(past_neighs, past_weights)])
								neigh_dico.update({'cumulated_presence_weighted': cum_sum_weighted + weights[n_index]})

						if symmetrize:
							setB.at[index_B[n_index], neigh_col] = sym_neigh

						neighs.append(neigh_dico)

					setA.at[index_A[k], neigh_col] = neighs

	return setA, setB


def compute_neighborhood_at_position(pos, distance, population=['targets', 'effectors'], theta_dist=None,
									 img_shape=(2048, 2048), return_tables=False, clear_neigh=False,
									 event_time_col=None,
									 neighborhood_kwargs={'mode': 'two-pop', 'status': None, 'not_status_option': None,
														  'include_dead_weight': True, "compute_cum_sum": False,
														  "attention_weight": True, 'symmetrize': True}):
	"""
	Computes neighborhood metrics for specified cell populations within a given position, based on distance criteria and additional parameters.

	This function assesses the neighborhood interactions between two specified cell populations (or within a single population) at a given position.
	It computes various neighborhood metrics based on specified distances, considering the entire image or excluding edge regions.
	The results are optionally cleared of previous neighborhood calculations and can be returned as updated tables.

	Parameters
	----------
	pos : str
		The path to the position directory where the analysis is to be performed.
	distance : float or list of float
		The distance(s) in pixels to define neighborhoods.
	population : list of str, optional
		Names of the cell populations to analyze. If a single population is provided, it is used for both populations in the analysis (default is ['targets', 'effectors']).
	theta_dist : float or list of float, optional
		Edge threshold(s) in pixels to exclude cells close to the image boundaries from the analysis. If not provided, defaults to 90% of each specified distance.
	img_shape : tuple of int, optional
		The dimensions (height, width) of the images in pixels (default is (2048, 2048)).
	return_tables : bool, optional
		If True, returns the updated data tables for both populations (default is False).
	clear_neigh : bool, optional
		If True, clears existing neighborhood columns from the data tables before computing new metrics (default is False).
	event_time_col : str, optional
		The column name indicating the event time for each cell, required if mean neighborhood metrics are to be computed before events.
	neighborhood_kwargs : dict, optional
		Additional keyword arguments for neighborhood computation, including mode, status options, and metrics (default includes mode 'two-pop', and symmetrization).

	Returns
	-------
	pandas.DataFrame or (pandas.DataFrame, pandas.DataFrame)
		If `return_tables` is True, returns the updated data tables for the specified populations. If only one population is analyzed, both returned data frames will be identical.

	Raises
	------
	AssertionError
		If the specified position path does not exist or if the number of distances and edge thresholds do not match.

	"""

	pos = pos.replace('\\', '/')
	pos = rf"{pos}"
	assert os.path.exists(pos), f'Position {pos} is not a valid path.'

	if isinstance(population, str):
		population = [population, population]

	if not isinstance(distance, list):
		distance = [distance]
	if not theta_dist is None and not isinstance(theta_dist, list):
		theta_dist = [theta_dist]

	if theta_dist is None:
		theta_dist = [0.9 * d for d in distance]
	assert len(theta_dist) == len(distance), 'Incompatible number of distances and number of edge thresholds.'

	if population[0] == population[1]:
		neighborhood_kwargs.update({'mode': 'self'})
	if population[1] != population[0]:
		neighborhood_kwargs.update({'mode': 'two-pop'})

	df_A, path_A = get_position_table(pos, population=population[0], return_path=True)
	df_B, path_B = get_position_table(pos, population=population[1], return_path=True)
	if df_A is None or df_B is None:
		return None

	if clear_neigh:
		if os.path.exists(path_A.replace('.csv','.pkl')):
			os.remove(path_A.replace('.csv','.pkl'))
		if os.path.exists(path_B.replace('.csv','.pkl')):
			os.remove(path_B.replace('.csv','.pkl'))
		df_pair, pair_path = get_position_table(pos, population='pairs', return_path=True)
		if df_pair is not None:
			os.remove(pair_path)


	df_A_pkl = get_position_pickle(pos, population=population[0], return_path=False)
	df_B_pkl = get_position_pickle(pos, population=population[1], return_path=False)

	if df_A_pkl is not None:
		pkl_columns = np.array(df_A_pkl.columns)
		neigh_columns = np.array([c.startswith('neighborhood') for c in pkl_columns])
		cols = list(pkl_columns[neigh_columns]) + ['FRAME']

		id_col = extract_identity_col(df_A_pkl)
		cols.append(id_col)
		on_cols = [id_col, 'FRAME']

		print(f'Recover {cols} from the pickle file...')
		try:
			df_A = pd.merge(df_A, df_A_pkl.loc[:,cols], how="outer", on=on_cols)
			print(df_A.columns)
		except Exception as e:
			print(f'Failure to merge pickle and csv files: {e}')

	if df_B_pkl is not None and df_B is not None:
		pkl_columns = np.array(df_B_pkl.columns)
		neigh_columns = np.array([c.startswith('neighborhood') for c in pkl_columns])
		cols = list(pkl_columns[neigh_columns]) + ['FRAME']

		id_col = extract_identity_col(df_B_pkl)
		cols.append(id_col)
		on_cols = [id_col, 'FRAME']

		print(f'Recover {cols} from the pickle file...')
		try:
			df_B = pd.merge(df_B, df_B_pkl.loc[:,cols], how="outer", on=on_cols)
		except Exception as e:
			print(f'Failure to merge pickle and csv files: {e}')

	if clear_neigh:
		unwanted = df_A.columns[df_A.columns.str.contains('neighborhood')]
		df_A = df_A.drop(columns=unwanted)
		unwanted = df_B.columns[df_B.columns.str.contains('neighborhood')]
		df_B = df_B.drop(columns=unwanted)

	df_A, df_B = distance_cut_neighborhood(df_A, df_B, distance, **neighborhood_kwargs)
	if df_A is None or df_B is None or len(df_A)==0:
		return None

	for td, d in zip(theta_dist, distance):

		if neighborhood_kwargs['mode'] == 'two-pop':
			neigh_col = f'neighborhood_2_circle_{d}_px'
		elif neighborhood_kwargs['mode'] == 'self':
			neigh_col = f'neighborhood_self_circle_{d}_px'

		# edge_filter_A = (df_A['POSITION_X'] > td)&(df_A['POSITION_Y'] > td)&(df_A['POSITION_Y'] < (img_shape[0] - td))&(df_A['POSITION_X'] < (img_shape[1] - td))
		# edge_filter_B = (df_B['POSITION_X'] > td)&(df_B['POSITION_Y'] > td)&(df_B['POSITION_Y'] < (img_shape[0] - td))&(df_B['POSITION_X'] < (img_shape[1] - td))
		# df_A.loc[~edge_filter_A, neigh_col] = np.nan
		# df_B.loc[~edge_filter_B, neigh_col] = np.nan

		print('Count neighborhood...')
		df_A = compute_neighborhood_metrics(df_A, neigh_col, metrics=['inclusive','exclusive','intermediate'], decompose_by_status=True)
		# if neighborhood_kwargs['symmetrize']:
		# 	df_B = compute_neighborhood_metrics(df_B, neigh_col, metrics=['inclusive','exclusive','intermediate'], decompose_by_status=True)
		print('Done...')

		if 'TRACK_ID' in list(df_A.columns):
			if not np.all(df_A['TRACK_ID'].isnull()):
				print('Estimate average neighborhood before/after event...')
				df_A = mean_neighborhood_before_event(df_A, neigh_col, event_time_col)
				if event_time_col is not None:
					df_A = mean_neighborhood_after_event(df_A, neigh_col, event_time_col)
				print('Done...')

	df_A.to_pickle(path_A.replace('.csv', '.pkl'))
	if not population[0] == population[1]:
		# Remove neighborhood column
		for td, d in zip(theta_dist, distance):
			if neighborhood_kwargs['mode'] == 'two-pop':
				neigh_col = f'neighborhood_2_circle_{d}_px'
			elif neighborhood_kwargs['mode'] == 'self':
				neigh_col = f'neighborhood_self_circle_{d}_px'
			df_B = df_B.drop(columns=[neigh_col])
		df_B.to_pickle(path_B.replace('.csv', '.pkl'))

	unwanted = df_A.columns[df_A.columns.str.startswith('neighborhood_')]
	df_A2 = df_A.drop(columns=unwanted)
	df_A2.to_csv(path_A, index=False)

	if not population[0] == population[1]:
		unwanted = df_B.columns[df_B.columns.str.startswith('neighborhood_')]
		df_B_csv = df_B.drop(unwanted, axis=1, inplace=False)
		df_B_csv.to_csv(path_B, index=False)

	if return_tables:
		return df_A, df_B

def compute_neighborhood_metrics(neigh_table, neigh_col, metrics=['inclusive','exclusive','intermediate'], decompose_by_status=False):

	"""
	Computes and appends neighborhood metrics to a dataframe based on specified neighborhood characteristics.

	This function iterates through a dataframe grouped by either 'TRACK_ID' or ['position', 'TRACK_ID'] (if 'position' column exists)
	and computes various neighborhood metrics (inclusive, exclusive, intermediate counts) for each cell. It can also decompose these
	metrics by cell status (e.g., live or dead) if specified.

	Parameters
	----------
	neigh_table : pandas.DataFrame
		A dataframe containing neighborhood information for each cell, including position, track ID, frame, and a specified neighborhood column.
	neigh_col : str
		The column name in `neigh_table` that contains neighborhood information (e.g., a list of neighbors with their attributes).
	metrics : list of str, optional
		The metrics to be computed from the neighborhood information. Possible values include 'inclusive', 'exclusive', and 'intermediate'.
		Default is ['inclusive', 'exclusive', 'intermediate'].
	decompose_by_status : bool, optional
		If True, the metrics are computed separately for different statuses (e.g., live or dead) of the neighboring cells. Default is False.

	Returns
	-------
	pandas.DataFrame
		The input dataframe with additional columns for each of the specified metrics, and, if `decompose_by_status` is True, separate
		metrics for each status.

	Notes
	-----
	- 'inclusive' count refers to the total number of neighbors.
	- 'exclusive' count refers to the number of neighbors that are closest.
	- 'intermediate' count refers to the sum of weights attributed to neighbors, representing a weighted count.
	- If `decompose_by_status` is True, metrics are appended with '_s0' or '_s1' to indicate the status they correspond to.

	Examples
	--------
	>>> neigh_table = pd.DataFrame({
	...     'TRACK_ID': [1, 1, 2, 2],
	...     'FRAME': [1, 2, 1, 2],
	...     'neighborhood_info': [{'weight': 1, 'status': 1, 'closest': 1}, ...]  # example neighborhood info
	... })
	>>> neigh_col = 'neighborhood_info'
	>>> updated_neigh_table = compute_neighborhood_metrics(neigh_table, neigh_col, metrics=['inclusive'], decompose_by_status=True)
	# Computes the inclusive count of neighbors for each cell, decomposed by cell status.

	"""

	neigh_table = neigh_table.reset_index(drop=True)
	if 'position' in list(neigh_table.columns):
		groupbycols = ['position']
	else:
		groupbycols = []

	id_col = extract_identity_col(neigh_table)
	groupbycols.append(id_col)

	neigh_table.sort_values(by=groupbycols+['FRAME'],inplace=True)

	for tid, group in neigh_table.groupby(groupbycols):
		group = group.dropna(subset=neigh_col)
		indices = list(group.index)
		neighbors = group[neigh_col].to_numpy()

		if 'inclusive' in metrics:
			n_inclusive = [len(n) for n in neighbors]

		if 'intermediate' in metrics:
			n_intermediate = np.zeros(len(neighbors))
			n_intermediate[:] = np.nan

		if 'exclusive' in metrics:
			n_exclusive = np.zeros(len(neighbors))
			n_exclusive[:] = np.nan

		if decompose_by_status:

			if 'inclusive' in metrics:
				n_inclusive_status_0 = np.zeros(len(neighbors))
				n_inclusive_status_0[:] = np.nan
				n_inclusive_status_1 = np.zeros(len(neighbors))
				n_inclusive_status_1[:] = np.nan

			if 'intermediate' in metrics:
				n_intermediate_status_0 = np.zeros(len(neighbors))
				n_intermediate_status_0[:] = np.nan
				n_intermediate_status_1 = np.zeros(len(neighbors))
				n_intermediate_status_1[:] = np.nan

			if 'exclusive' in metrics:
				n_exclusive_status_0 = np.zeros(len(neighbors))
				n_exclusive_status_0[:] = np.nan
				n_exclusive_status_1 = np.zeros(len(neighbors))
				n_exclusive_status_1[:] = np.nan

		for t in range(len(neighbors)):

			neighs_at_t = neighbors[t]
			weights_at_t = [n['weight'] for n in neighs_at_t]
			status_at_t = [n['status'] for n in neighs_at_t]
			closest_at_t = [n['closest'] for n in neighs_at_t]

			if 'intermediate' in metrics:
				n_intermediate[t] = np.sum(weights_at_t)
			if 'exclusive' in metrics:
				n_exclusive[t] = sum([c == 1.0 for c in closest_at_t])

			if decompose_by_status:

				if 'inclusive' in metrics:
					n_inclusive_status_0[t] = sum([s == 0.0 for s in status_at_t])
					n_inclusive_status_1[t] = sum([s == 1.0 for s in status_at_t])

				if 'intermediate' in metrics:
					weights_at_t = np.array(weights_at_t)

					# intermediate
					weights_status_1 = weights_at_t[np.array([s == 1.0 for s in status_at_t], dtype=bool)]
					weights_status_0 = weights_at_t[np.array([s == 0.0 for s in status_at_t], dtype=bool)]
					n_intermediate_status_1[t] = np.sum(weights_status_1)
					n_intermediate_status_0[t] = np.sum(weights_status_0)

				if 'exclusive' in metrics:
					n_exclusive_status_0[t] = sum(
						[c == 1.0 if s == 0.0 else False for c, s in zip(closest_at_t, status_at_t)])
					n_exclusive_status_1[t] = sum(
						[c == 1.0 if s == 1.0 else False for c, s in zip(closest_at_t, status_at_t)])

		if 'inclusive' in metrics:
			neigh_table.loc[indices, 'inclusive_count_' + neigh_col] = n_inclusive
		if 'intermediate' in metrics:
			neigh_table.loc[indices, 'intermediate_count_' + neigh_col] = n_intermediate
		if 'exclusive' in metrics:
			neigh_table.loc[indices, 'exclusive_count_' + neigh_col] = n_exclusive

		if decompose_by_status:
			if 'inclusive' in metrics:
				neigh_table.loc[indices, 'inclusive_count_s0_' + neigh_col] = n_inclusive_status_0
				neigh_table.loc[indices, 'inclusive_count_s1_' + neigh_col] = n_inclusive_status_1
			if 'intermediate' in metrics:
				neigh_table.loc[indices, 'intermediate_count_s0_' + neigh_col] = n_intermediate_status_0
				neigh_table.loc[indices, 'intermediate_count_s1_' + neigh_col] = n_intermediate_status_1
			if 'exclusive' in metrics:
				neigh_table.loc[indices, 'exclusive_count_s0_' + neigh_col] = n_exclusive_status_0
				neigh_table.loc[indices, 'exclusive_count_s1_' + neigh_col] = n_exclusive_status_1

	return neigh_table


def mean_neighborhood_before_event(neigh_table, neigh_col, event_time_col,
								   metrics=['inclusive', 'exclusive', 'intermediate']):
	"""
	Computes the mean neighborhood metrics for each cell track before a specified event time.

	This function calculates the mean values of specified neighborhood metrics (inclusive, exclusive, intermediate)
	for each cell track up to and including the frame of an event. The function requires the neighborhood metrics to
	have been previously computed and appended to the input dataframe. It operates on grouped data based on position
	and track ID, handling cases with or without position information.

	Parameters
	----------
	neigh_table : pandas.DataFrame
		A dataframe containing cell track data with precomputed neighborhood metrics and event time information.
	neigh_col : str
		The base name of the neighborhood metric columns in `neigh_table`.
	event_time_col : str or None
		The column name indicating the event time for each cell track. If None, the maximum frame number in the
		dataframe is used as the event time for all tracks.

	Returns
	-------
	pandas.DataFrame
		The input dataframe with added columns for the mean neighborhood metrics before the event for each cell track.
		The new columns are named as 'mean_count_{metric}_{neigh_col}_before_event', where {metric} is one of
		'inclusive', 'exclusive', 'intermediate'.

	"""


	neigh_table = neigh_table.reset_index(drop=True)
	if 'position' in list(neigh_table.columns):
		groupbycols = ['position']
	else:
		groupbycols = []

	id_col = extract_identity_col(neigh_table)
	groupbycols.append(id_col)

	neigh_table.sort_values(by=groupbycols+['FRAME'],inplace=True)
	suffix = '_before_event'

	if event_time_col is None:
		print('No event time was provided... Estimating the mean neighborhood over the whole observation time...')
		neigh_table.loc[:, 'event_time_temp'] = neigh_table['FRAME'].max()
		event_time_col = 'event_time_temp'
		suffix = ''

	for tid, group in neigh_table.groupby(groupbycols):

		group = group.dropna(subset=neigh_col)
		indices = list(group.index)

		event_time_values = group[event_time_col].to_numpy()
		if len(event_time_values) > 0:
			event_time = event_time_values[0]
		else:
			continue

		if event_time < 0.:
			event_time = group['FRAME'].max()

		if 'intermediate' in metrics:
			valid_counts_intermediate = group.loc[
				group['FRAME'] <= event_time, 'intermediate_count_s1_' + neigh_col].to_numpy()
			if len(valid_counts_intermediate[valid_counts_intermediate == valid_counts_intermediate]) > 0:
				neigh_table.loc[indices, f'mean_count_intermediate_{neigh_col}{suffix}'] = np.nanmean(
					valid_counts_intermediate)
		if 'inclusive' in metrics:
			valid_counts_inclusive = group.loc[
				group['FRAME'] <= event_time, 'inclusive_count_s1_' + neigh_col].to_numpy()
			if len(valid_counts_inclusive[valid_counts_inclusive == valid_counts_inclusive]) > 0:
				neigh_table.loc[indices, f'mean_count_inclusive_{neigh_col}{suffix}'] = np.nanmean(
					valid_counts_inclusive)
		if 'exclusive' in metrics:
			valid_counts_exclusive = group.loc[
				group['FRAME'] <= event_time, 'exclusive_count_s1_' + neigh_col].to_numpy()
			if len(valid_counts_exclusive[valid_counts_exclusive == valid_counts_exclusive]) > 0:
				neigh_table.loc[indices, f'mean_count_exclusive_{neigh_col}{suffix}'] = np.nanmean(
					valid_counts_exclusive)

	if event_time_col == 'event_time_temp':
		neigh_table = neigh_table.drop(columns='event_time_temp')
	return neigh_table


def mean_neighborhood_after_event(neigh_table, neigh_col, event_time_col,
								  metrics=['inclusive', 'exclusive', 'intermediate']):
	"""
	Computes the mean neighborhood metrics for each cell track after a specified event time.

	This function calculates the mean values of specified neighborhood metrics (inclusive, exclusive, intermediate)
	for each cell track after the event time. The function requires the neighborhood metrics to
	have been previously computed and appended to the input dataframe. It operates on grouped data based on position
	and track ID, handling cases with or without position information.

	Parameters
	----------
	neigh_table : pandas.DataFrame
		A dataframe containing cell track data with precomputed neighborhood metrics and event time information.
	neigh_col : str
		The base name of the neighborhood metric columns in `neigh_table`.
	event_time_col : str or None
		The column name indicating the event time for each cell track. If None, the maximum frame number in the
		dataframe is used as the event time for all tracks.

	Returns
	-------
	pandas.DataFrame
		The input dataframe with added columns for the mean neighborhood metrics before the event for each cell track.
		The new columns are named as 'mean_count_{metric}_{neigh_col}_before_event', where {metric} is one of
		'inclusive', 'exclusive', 'intermediate'.

	"""


	neigh_table = neigh_table.reset_index(drop=True)
	if 'position' in list(neigh_table.columns):
		groupbycols = ['position']
	else:
		groupbycols = []

	id_col = extract_identity_col(neigh_table)
	groupbycols.append(id_col)

	neigh_table.sort_values(by=groupbycols+['FRAME'],inplace=True)
	suffix = '_after_event'

	if event_time_col is None:
		neigh_table.loc[:, 'event_time_temp'] = None  # neigh_table['FRAME'].max()
		event_time_col = 'event_time_temp'
		suffix = ''

	for tid, group in neigh_table.groupby(groupbycols):

		group = group.dropna(subset=neigh_col)
		indices = list(group.index)

		event_time_values = group[event_time_col].to_numpy()
		if len(event_time_values) > 0:
			event_time = event_time_values[0]
		else:
			continue

		if event_time is not None and (event_time >= 0.):

			if 'intermediate' in metrics:
				valid_counts_intermediate = group.loc[
					group['FRAME'] > event_time, 'intermediate_count_s1_' + neigh_col].to_numpy()
				if len(valid_counts_intermediate[valid_counts_intermediate == valid_counts_intermediate]) > 0:
					neigh_table.loc[indices, f'mean_count_intermediate_{neigh_col}{suffix}'] = np.nanmean(
						valid_counts_intermediate)
			if 'inclusive' in metrics:
				valid_counts_inclusive = group.loc[
					group['FRAME'] > event_time, 'inclusive_count_s1_' + neigh_col].to_numpy()
				if len(valid_counts_inclusive[valid_counts_inclusive == valid_counts_inclusive]) > 0:
					neigh_table.loc[indices, f'mean_count_inclusive_{neigh_col}{suffix}'] = np.nanmean(
						valid_counts_inclusive)
			if 'exclusive' in metrics:
				valid_counts_exclusive = group.loc[
					group['FRAME'] > event_time, 'exclusive_count_s1_' + neigh_col].to_numpy()
				if len(valid_counts_exclusive[valid_counts_exclusive == valid_counts_exclusive]) > 0:
					neigh_table.loc[indices, f'mean_count_exclusive_{neigh_col}{suffix}'] = np.nanmean(
						valid_counts_exclusive)

	if event_time_col == 'event_time_temp':
		neigh_table = neigh_table.drop(columns='event_time_temp')

	return neigh_table


# New functions for direct cell-cell contact neighborhood

def sign(num):
	return -1 if num < 0 else 1


def contact_neighborhood(labelsA, labelsB=None, border=3, connectivity=2):

	labelsA = labelsA.astype(float)
	if labelsB is not None:
		labelsB = labelsB.astype(float)

	print(f"Border = {border}")

	if border > 0:
		print(labelsA.shape, border * (-1))
		labelsA_edge = contour_of_instance_segmentation(label=labelsA, distance=border * (-1)).astype(float)
		labelsA[np.where(labelsA_edge > 0)] = labelsA_edge[np.where(labelsA_edge > 0)]
		if labelsB is not None:
			labelsB_edge = contour_of_instance_segmentation(label=labelsB, distance=border * (-1)).astype(float)
			labelsB[np.where(labelsB_edge > 0)] = labelsB_edge[np.where(labelsB_edge > 0)]

	if labelsB is not None:
		labelsA[labelsA != 0] = -labelsA[labelsA != 0]
		labelsAB = merge_labels(labelsA, labelsB)
		labelsBA = merge_labels(labelsB, labelsA)
		label_cases = [labelsAB, labelsBA]
	else:
		label_cases = [labelsA]

	coocurrences = []
	for lbl in label_cases:
		coocurrences.extend(find_contact_neighbors(lbl, connectivity=connectivity))

	unique_pairs = np.unique(coocurrences, axis=0)

	if labelsB is not None:
		neighs = np.unique([tuple(sorted(p)) for p in unique_pairs if p[0] != p[1] and sign(p[0]) != sign(p[1])],
						   axis=0)
	else:
		neighs = np.unique([tuple(sorted(p)) for p in unique_pairs if p[0] != p[1]], axis=0)

	return neighs

def merge_labels(labelsA, labelsB):
	labelsA = labelsA.astype(float)
	labelsB = labelsB.astype(float)

	labelsAB = labelsA.copy()
	labelsAB[np.where(labelsB != 0)] = labelsB[np.where(labelsB != 0)]

	return labelsAB


def find_contact_neighbors(labels, connectivity=2):
	assert labels.ndim == 2, "Wrong dimension for labels..."
	g, nodes = pixel_graph(labels, mask=labels.astype(bool), connectivity=connectivity)
	g.eliminate_zeros()

	coo = g.tocoo()
	center_coords = nodes[coo.row]
	neighbor_coords = nodes[coo.col]

	center_values = labels.ravel()[center_coords]
	neighbor_values = labels.ravel()[neighbor_coords]
	touching_masks = np.column_stack((center_values, neighbor_values))

	return touching_masks


def mask_contact_neighborhood(setA, setB, labelsA, labelsB, distance, mode='two-pop', status=None,
							  not_status_option=None, compute_cum_sum=True,
							  attention_weight=True, symmetrize=True, include_dead_weight=True,
							  column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y',
											 'mask_id': 'class_id'}):
	"""

	Match neighbors in set A and B within a circle of radius d.

	Parameters
	----------
	setA,setB : pandas DataFrame
		Trajectory or position sets A and B.
	distance : float
		Cut-distance in pixels to match neighboring pairs.
	mode: str
		neighboring mode, between 'two-pop' (e.g. target-effector) and 'self' (target-target or effector-effector).
	status: None or status
		name to look for cells to ignore (because they are dead). By default all cells are kept.
	compute_cum_sum: bool,
		compute cumulated time of presence of neighbours (only if trajectories available for both sets)
	attention_weight: bool,
		compute the attention weight (how much a cell of set B is shared across cells of set A)
	symmetrize: bool,
		write in set B the neighborhood of set A
	include_dead_weight: bool
		do not count dead cells when establishing attention weight
	"""

	# Check live_status option
	if setA is not None and setB is not None:
		setA, setB, status = set_live_status(setA, setB, status, not_status_option)
	else:
		return None, None

	# Check distance option
	if not isinstance(distance, list):
		distance = [distance]

	for d in distance:
		# loop over each provided distance

		if mode == 'two-pop':
			neigh_col = f'neighborhood_2_contact_{d}_px'
		elif mode == 'self':
			neigh_col = f'neighborhood_self_contact_{d}_px'

		cl = []
		for s in [setA, setB]:

			# Check whether data can be tracked
			temp_column_labels = column_labels.copy()

			id_col = extract_identity_col(s)
			temp_column_labels.update({'track': id_col})
			if id_col=='ID':
				compute_cum_sum = False  # if no tracking data then cum_sum is not relevant
			cl.append(temp_column_labels)

			# Remove nan tracks (cells that do not belong to a track)
			s[neigh_col] = np.nan
			s[neigh_col] = s[neigh_col].astype(object)
			s.dropna(subset=[cl[-1]['track']], inplace=True)

		# Loop over each available timestep
		timeline = np.unique(np.concatenate([setA[cl[0]['time']].to_numpy(), setB[cl[1]['time']].to_numpy()])).astype(
			int)
		for t in tqdm(timeline):

			index_A = list(setA.loc[setA[cl[0]['time']] == t].index)
			coordinates_A = setA.loc[setA[cl[0]['time']] == t, [cl[0]['x'], cl[0]['y']]].to_numpy()
			ids_A = setA.loc[setA[cl[0]['time']] == t, cl[0]['track']].to_numpy()
			mask_ids_A = setA.loc[setA[cl[0]['time']] == t, cl[0]['mask_id']].to_numpy()
			status_A = setA.loc[setA[cl[0]['time']] == t, status[0]].to_numpy()

			index_B = list(setB.loc[setB[cl[1]['time']] == t].index)
			coordinates_B = setB.loc[setB[cl[1]['time']] == t, [cl[1]['x'], cl[1]['y']]].to_numpy()
			ids_B = setB.loc[setB[cl[1]['time']] == t, cl[1]['track']].to_numpy()
			mask_ids_B = setB.loc[setB[cl[1]['time']] == t, cl[1]['mask_id']].to_numpy()
			status_B = setB.loc[setB[cl[1]['time']] == t, status[1]].to_numpy()

			print(f"Frame {t}")
			print(f"{mask_ids_A=}", f"{mask_ids_B}")

			if len(ids_A) > 0 and len(ids_B) > 0:

				# compute distance matrix
				dist_map = cdist(coordinates_A, coordinates_B, metric="euclidean")

				# Do the mask contact computation
				if labelsB is not None:
					lblB = labelsB[t]
				else:
					lblB = labelsB

				print(f"Distance {d} for contact as border")
				contact_pairs = contact_neighborhood(labelsA[t], labelsB=lblB, border=d, connectivity=2)

				print(t, f"{np.unique(labelsA[t])=}")
				print(f"Frame {t}: found the following contact pairs: {contact_pairs}...")
				# Put infinite distance to all non-contact pairs (something like this)
				plot_map = False

				if len(contact_pairs) > 0:
					mask = np.ones_like(dist_map).astype(bool)

					indices_to_keep = []
					for cp in contact_pairs:

						if np.any(cp < 0):
							if cp[0] < 0:
								mask_A = cp[1]
								mask_B = np.abs(cp[0])
							else:
								mask_A = cp[0]
								mask_B = np.abs(cp[1])
						else:
							mask_A = cp[0]
							mask_B = cp[1]

						try:

							idx_A = np.where(mask_ids_A == int(mask_A))[0][0]
							idx_B = np.where(mask_ids_B == int(mask_B))[0][0]
							print(idx_A, idx_B)
							indices_to_keep.append([idx_A,idx_B])
						except Exception as e:
							print(f'{e} {t=} error something happened!!')
							pass

					print(f'Indices to keep: {indices_to_keep}...')
					if len(indices_to_keep) > 0:
						indices_to_keep = np.array(indices_to_keep)
						mask[indices_to_keep[:, 0], indices_to_keep[:, 1]] = False
						if mode == 'self':
							mask[indices_to_keep[:, 1], indices_to_keep[:, 0]] = False
						dist_map[mask] = 1.0E06
						plot_map=True
					else:
						dist_map[:,:] = 1.0E06
				else:
					dist_map[:, :] = 1.0E06

				# PROCEED all the same?? --> I guess so
				# if plot_map:
				# 	import matplotlib.pyplot as plt
				# 	print(indices_to_keep)
				# 	plt.imshow(dist_map)
				# 	plt.pause(5)
				# 	plt.close()

				d_filter = 1.0E05
				if attention_weight:
					weights, closest_A = compute_attention_weight(dist_map, d_filter, status_A, ids_A, axis=1,
																  include_dead_weight=include_dead_weight)

				# Target centric
				for k in range(dist_map.shape[0]):

					col = dist_map[k, :]
					col[col == 0.] = 1.0E06

					neighs_B = np.array([ids_B[i] for i in np.where((col <= d_filter))[0]])
					status_neigh_B = np.array([status_B[i] for i in np.where((col <= d_filter))[0]])
					dist_B = [round(col[i], 2) for i in np.where((col <= d_filter))[0]]
					if len(dist_B) > 0:
						closest_B_cell = neighs_B[np.argmin(dist_B)]

					if symmetrize and attention_weight:
						n_neighs = float(len(neighs_B))
						if not include_dead_weight:
							n_neighs_alive = len(np.where(status_neigh_B == 1)[0])
							neigh_count = n_neighs_alive
						else:
							neigh_count = n_neighs
						if neigh_count > 0:
							weight_A = 1. / neigh_count
						else:
							weight_A = np.nan

						if not include_dead_weight and status_A[k] == 0:
							weight_A = 0

					neighs = []
					setA.at[index_A[k], neigh_col] = []
					for n in range(len(neighs_B)):

						# index in setB
						n_index = np.where(ids_B == neighs_B[n])[0][0]
						# Assess if neigh B is closest to A
						if attention_weight:
							if closest_A[n_index] == ids_A[k]:
								closest = True
							else:
								closest = False

						if symmetrize:
							# Load neighborhood previous data
							sym_neigh = setB.loc[index_B[n_index], neigh_col]
							if neighs_B[n] == closest_B_cell:
								closest_b = True
							else:
								closest_b = False
							if isinstance(sym_neigh, list):
								sym_neigh.append({'id': ids_A[k], 'distance': dist_B[n], 'status': status_A[k]})
							else:
								sym_neigh = [{'id': ids_A[k], 'distance': dist_B[n], 'status': status_A[k]}]
							if attention_weight:
								sym_neigh[-1].update({'weight': weight_A, 'closest': closest_b})

						# Write the minimum info about neighborhing cell B
						neigh_dico = {'id': neighs_B[n], 'distance': dist_B[n], 'status': status_neigh_B[n]}
						if attention_weight:
							neigh_dico.update({'weight': weights[n_index], 'closest': closest})

						if compute_cum_sum:
							# Compute the integrated presence of the neighboring cell B
							assert cl[1][
									   'track'] == 'TRACK_ID', 'The set B does not seem to contain tracked data. The cumulative time will be meaningless.'
							past_neighs = [[ll['id'] for ll in l] if len(l) > 0 else [None] for l in setA.loc[
								(setA[cl[0]['track']] == ids_A[k]) & (setA[cl[0]['time']] <= t), neigh_col].to_numpy()]
							past_neighs = [item for sublist in past_neighs for item in sublist]

							if attention_weight:
								past_weights = [[ll['weight'] for ll in l] if len(l) > 0 else [None] for l in setA.loc[
									(setA[cl[0]['track']] == ids_A[k]) & (
												setA[cl[0]['time']] <= t), neigh_col].to_numpy()]
								past_weights = [item for sublist in past_weights for item in sublist]

							cum_sum = len(np.where(past_neighs == neighs_B[n])[0])
							neigh_dico.update({'cumulated_presence': cum_sum + 1})

							if attention_weight:
								cum_sum_weighted = np.sum(
									[w if l == neighs_B[n] else 0 for l, w in zip(past_neighs, past_weights)])
								neigh_dico.update({'cumulated_presence_weighted': cum_sum_weighted + weights[n_index]})

						if symmetrize:
							setB.at[index_B[n_index], neigh_col] = sym_neigh

						neighs.append(neigh_dico)

					setA.at[index_A[k], neigh_col] = neighs

	return setA, setB


def compute_contact_neighborhood_at_position(pos, distance, population=['targets', 'effectors'], theta_dist=None,
											 img_shape=(2048, 2048), return_tables=False, clear_neigh=False,
											 event_time_col=None,
											 neighborhood_kwargs={'mode': 'two-pop', 'status': None,
																  'not_status_option': None,
																  'include_dead_weight': True, "compute_cum_sum": False,
																  "attention_weight": True, 'symmetrize': True}):
	"""
	Computes neighborhood metrics for specified cell populations within a given position, based on distance criteria and additional parameters.

	This function assesses the neighborhood interactions between two specified cell populations (or within a single population) at a given position.
	It computes various neighborhood metrics based on specified distances, considering the entire image or excluding edge regions.
	The results are optionally cleared of previous neighborhood calculations and can be returned as updated tables.

	Parameters
	----------
	pos : str
		The path to the position directory where the analysis is to be performed.
	distance : float or list of float
		The distance(s) in pixels to define neighborhoods.
	population : list of str, optional
		Names of the cell populations to analyze. If a single population is provided, it is used for both populations in the analysis (default is ['targets', 'effectors']).
	theta_dist : float or list of float, optional
		Edge threshold(s) in pixels to exclude cells close to the image boundaries from the analysis. If not provided, defaults to 90% of each specified distance.
	img_shape : tuple of int, optional
		The dimensions (height, width) of the images in pixels (default is (2048, 2048)).
	return_tables : bool, optional
		If True, returns the updated data tables for both populations (default is False).
	clear_neigh : bool, optional
		If True, clears existing neighborhood columns from the data tables before computing new metrics (default is False).
	event_time_col : str, optional
		The column name indicating the event time for each cell, required if mean neighborhood metrics are to be computed before events.
	neighborhood_kwargs : dict, optional
		Additional keyword arguments for neighborhood computation, including mode, status options, and metrics (default includes mode 'two-pop', and symmetrization).

	Returns
	-------
	pandas.DataFrame or (pandas.DataFrame, pandas.DataFrame)
		If `return_tables` is True, returns the updated data tables for the specified populations. If only one population is analyzed, both returned data frames will be identical.

	Raises
	------
	AssertionError
		If the specified position path does not exist or if the number of distances and edge thresholds do not match.

	"""

	pos = pos.replace('\\', '/')
	pos = rf"{pos}"
	assert os.path.exists(pos), f'Position {pos} is not a valid path.'

	if isinstance(population, str):
		population = [population, population]

	if not isinstance(distance, list):
		distance = [distance]
	if not theta_dist is None and not isinstance(theta_dist, list):
		theta_dist = [theta_dist]

	if theta_dist is None:
		theta_dist = [0 for d in distance]  # 0.9*d
	assert len(theta_dist) == len(distance), 'Incompatible number of distances and number of edge thresholds.'

	if population[0] == population[1]:
		neighborhood_kwargs.update({'mode': 'self'})
	if population[1] != population[0]:
		neighborhood_kwargs.update({'mode': 'two-pop'})

	df_A, path_A = get_position_table(pos, population=population[0], return_path=True)
	df_B, path_B = get_position_table(pos, population=population[1], return_path=True)
	if df_A is None or df_B is None:
		return None

	if clear_neigh:
		if os.path.exists(path_A.replace('.csv','.pkl')):
			os.remove(path_A.replace('.csv','.pkl'))
		if os.path.exists(path_B.replace('.csv','.pkl')):
			os.remove(path_B.replace('.csv','.pkl'))
		df_pair, pair_path = get_position_table(pos, population='pairs', return_path=True)
		if df_pair is not None:
			os.remove(pair_path)

	df_A_pkl = get_position_pickle(pos, population=population[0], return_path=False)
	df_B_pkl = get_position_pickle(pos, population=population[1], return_path=False)

	if df_A_pkl is not None:
		pkl_columns = np.array(df_A_pkl.columns)
		neigh_columns = np.array([c.startswith('neighborhood') for c in pkl_columns])
		cols = list(pkl_columns[neigh_columns]) + ['FRAME']
		
		id_col = extract_identity_col(df_A_pkl)
		cols.append(id_col)
		on_cols = [id_col, 'FRAME']

		print(f'Recover {cols} from the pickle file...')
		try:
			df_A = pd.merge(df_A, df_A_pkl.loc[:,cols], how="outer", on=on_cols)
			print(df_A.columns)
		except Exception as e:
			print(f'Failure to merge pickle and csv files: {e}')

	if df_B_pkl is not None and df_B is not None:
		pkl_columns = np.array(df_B_pkl.columns)
		neigh_columns = np.array([c.startswith('neighborhood') for c in pkl_columns])
		cols = list(pkl_columns[neigh_columns]) + ['FRAME']

		id_col = extract_identity_col(df_B_pkl)
		cols.append(id_col)
		on_cols = [id_col, 'FRAME']

		print(f'Recover {cols} from the pickle file...')
		try:
			df_B = pd.merge(df_B, df_B_pkl.loc[:,cols], how="outer", on=on_cols)
		except Exception as e:
			print(f'Failure to merge pickle and csv files: {e}')

	labelsA = locate_labels(pos, population=population[0])
	if population[1] == population[0]:
		labelsB = None
	else:
		labelsB = locate_labels(pos, population=population[1])

	if clear_neigh:
		unwanted = df_A.columns[df_A.columns.str.contains('neighborhood')]
		df_A = df_A.drop(columns=unwanted)
		unwanted = df_B.columns[df_B.columns.str.contains('neighborhood')]
		df_B = df_B.drop(columns=unwanted)

	print(f"Distance: {distance} for mask contact")
	df_A, df_B = mask_contact_neighborhood(df_A, df_B, labelsA, labelsB, distance, **neighborhood_kwargs)
	if df_A is None or df_B is None or len(df_A)==0:
		return None

	for td, d in zip(theta_dist, distance):

		if neighborhood_kwargs['mode'] == 'two-pop':
			neigh_col = f'neighborhood_2_contact_{d}_px'
		elif neighborhood_kwargs['mode'] == 'self':
			neigh_col = f'neighborhood_self_contact_{d}_px'

		df_A.loc[df_A['class_id'].isnull(),neigh_col] = np.nan

		# edge_filter_A = (df_A['POSITION_X'] > td)&(df_A['POSITION_Y'] > td)&(df_A['POSITION_Y'] < (img_shape[0] - td))&(df_A['POSITION_X'] < (img_shape[1] - td))
		# edge_filter_B = (df_B['POSITION_X'] > td)&(df_B['POSITION_Y'] > td)&(df_B['POSITION_Y'] < (img_shape[0] - td))&(df_B['POSITION_X'] < (img_shape[1] - td))
		# df_A.loc[~edge_filter_A, neigh_col] = np.nan
		# df_B.loc[~edge_filter_B, neigh_col] = np.nan

		df_A = compute_neighborhood_metrics(df_A, neigh_col, metrics=['inclusive', 'intermediate'],
											decompose_by_status=True)
		if 'TRACK_ID' in list(df_A.columns):
			if not np.all(df_A['TRACK_ID'].isnull()):
				df_A = mean_neighborhood_before_event(df_A, neigh_col, event_time_col, metrics=['inclusive','intermediate'])
				if event_time_col is not None:
					df_A = mean_neighborhood_after_event(df_A, neigh_col, event_time_col, metrics=['inclusive', 'intermediate'])
				print('Done...')
				
	df_A.to_pickle(path_A.replace('.csv', '.pkl'))
	if not population[0] == population[1]:
		# Remove neighborhood column
		for td, d in zip(theta_dist, distance):
			if neighborhood_kwargs['mode'] == 'two-pop':
				neigh_col = f'neighborhood_2_contact_{d}_px'
			elif neighborhood_kwargs['mode'] == 'self':
				neigh_col = f'neighborhood_self_contact_{d}_px'
			df_B = df_B.drop(columns=[neigh_col])
		df_B.to_pickle(path_B.replace('.csv', '.pkl'))

	unwanted = df_A.columns[df_A.columns.str.startswith('neighborhood_')]
	df_A2 = df_A.drop(columns=unwanted)
	df_A2.to_csv(path_A, index=False)

	if not population[0] == population[1]:
		unwanted = df_B.columns[df_B.columns.str.startswith('neighborhood_')]
		df_B_csv = df_B.drop(unwanted, axis=1, inplace=False)
		df_B_csv.to_csv(path_B, index=False)

	if return_tables:
		return df_A, df_B



# def mask_intersection_neighborhood(setA, labelsA, setB, labelsB, threshold_iou=0.5, viewpoint='B'):
# 	# do whatever to match objects in A and B
# 	return setA, setB

if __name__ == "__main__":

	print('None')
	pos = "/home/torro/Documents/Experiments/NKratio_Exp/W5/500"

	test, _ = compute_neighborhood_at_position(pos, [62], population=['targets', 'effectors'], theta_dist=None,
											   img_shape=(2048, 2048), return_tables=True, clear_neigh=True,
											   neighborhood_kwargs={'mode': 'two-pop', 'status': ['class', None],
																	'not_status_option': [True, False],
																	'include_dead_weight': True,
																	"compute_cum_sum": False, "attention_weight": True,
																	'symmetrize': False})

	# test = compute_neighborhood_metrics(test, 'neighborhood_self_circle_150_px', metrics=['inclusive','exclusive','intermediate'], decompose_by_status=True)
	print(test.columns)
	#print(segment(None,'test'))



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt
from celldetective.utils import get_software_location
import os
from celldetective.gui.gui_utils import center_window

class AboutWidget(QWidget):

	def __init__(self):

		super().__init__()
		self.setWindowTitle("About celldetective")
		self.setMinimumWidth(300)
		center_window(self)
		logo = QPixmap(os.sep.join([get_software_location(),'celldetective','icons','logo.png']))

		# Create the layout
		layout = QVBoxLayout(self)
		img_label = QLabel('')
		img_label.setPixmap(logo)
		layout.addWidget(img_label, alignment=Qt.AlignCenter)

		self.soft_name = QLabel('celldetective')
		self.soft_name.setStyleSheet("""font-weight: bold;
										font-size: 18px;
									""")
		layout.addWidget(self.soft_name, alignment=Qt.AlignCenter)

		self.version_lbl = QLabel(f"Version X.X.X <a href=\"https://github.com/remyeltorro/celldetective/releases\">(release notes)</a>")
		self.version_lbl.setOpenExternalLinks(True)
		layout.addWidget(self.version_lbl, alignment=Qt.AlignCenter)

		self.lab_lbl = QLabel("Developed at Laboratoire Adhésion et Inflammation (LAI) INSERM U1067 CNRS UMR 7333")
		self.lab_lbl.setWordWrap(True)
		layout.addWidget(self.lab_lbl, alignment=Qt.AlignCenter)

		self.centuri_mention = QLabel("The project leading to this publication has received funding from France 2030, the French Government program managed by the French National Research Agency (ANR-16-CONV-0001) and from Excellence Initiative of Aix-Marseille University - A*MIDEX')")
		self.centuri_mention.setWordWrap(True)
		layout.addWidget(self.centuri_mention, alignment=Qt.AlignCenter)



